
from epyk.core.html import Html

from epyk.core.js import JsUtils
from epyk.core.js.html import JsHtmlStepper
from epyk.core.html.options import OptPanel


class Step(object):

  def __init__(self, src, selector):
    self._src = src
    self._selector = selector

  def click(self, jsFncs, profile=False):

    if not isinstance(jsFncs, list):
      jsFncs = [jsFncs]
    self._src._report._props['js'].setdefault("onReady", []).append("%s.addEventListener('click', function(event){%s})" % (self._selector.varName, JsUtils.jsConvertFncs(jsFncs, toStr=True)))
    return self


class Stepper(Html.Html):

  def __init__(self, report, records, width, height, color, options):
    super(Stepper, self).__init__(report, records, css_attrs={"list-style-type": 'none', "width": width})
    self.color = self._report.theme.greys[-1] if color is None else color
    self.css({'color': self.color, "margin": '0', 'display': 'block'})
    dflt_options = {'svg_style': {'display': 'block', 'width': 100, 'height': height[0]-20}, 'circle_factor': 2,
                    'text_style': {'display': 'block', 'text-align': 'center'},
                    'backgrounds': {"success": '#37A78C', 'error': '#FF0000', 'waiting': '#A0A0A0', 'pending': '#FF9200'},
                    'success': ["#C9EDE4", "#63CBB2", "#37A78C"],
                    'error': ["#F8CBAD", "#FF5757", "#FF0000"],
                    'pending': ["#FFDEB3", "#FFB047", "#FF9200"],
                    'waiting': ["#BEBEBE", "#B5B5B5", "#A0A0A0"],
                    'shape': 'circle', 'text_colors': 'white'}
    dflt_options.update(options)
    self.__options = OptPanel.OptionsStepper(self, dflt_options)

  def __getitem__(self, i):
    """
    Description:
    ------------

    """
    return Step(self, selector=self.dom[i])

  @property
  def dom(self):
    """
    Description:
    ------------

    :rtype: JsHtmlStepper.Stepper
    """
    if self._dom is None:
      self._dom = JsHtmlStepper.Stepper(self, report=self._report)
    return self._dom

  @property
  def options(self):
    """
    Description:
    ------------
    Property to set all the possible object for a button

    :rtype: OptPanel.OptionsStepper
    """
    return self.__options

  @property
  def _js__builder__(self):
    # li.style.float = "left";
    return ''' htmlObj.innerHTML = '';
      var width = options.svg_style.width; var height = options.svg_style.height;
      var attrs = ['name', 'text', 'title', 'tooltip'];
      var props = ['color', 'background', 'shape', 'title_color'];
      
      data.forEach(function(step, i){
        var li = document.createElement("LI") ;
        li.style['margin-bottom'] = '10px';
        
        attrs.forEach(function(attr){ if (typeof step[attr] === 'undefined'){ step[attr] = ''}; });
        props.forEach(function(prop){ if (typeof step[prop] === 'undefined'){ step[prop] = options[prop]}; })

        var span = document.createElement("SPAN"); span.setAttribute('name', step.name);
        span.setAttribute('name', 'label'); span.innerHTML = step.title;
        var div = document.createElement("DIV"); 
        div.setAttribute('title', step.tooltip);
        div.setAttribute('name', 'svg_holder');
        
        for (var key in options.text_style){ span.style[key] = options.text_style[key] }
        span.style.width = width +"px";

        div.appendChild(span); li.appendChild(div);
        div.id = "svg_" + i;
        window[step.shape](div, options, step.status); 
        htmlObj.appendChild(li)
      })    
      '''

  def add_step(self):
    """"""

  def add_shape(self, shape, shape_def):
    """

    :param shape:
    :param shape_def:
    """
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    constructors[shape] = "function %s(htmlObj, options, status){%s}" % (shape, JsHtmlStepper.JsShapes().custom(shape_def))
    return self

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    # add all the shape definitions
    constructors = self._report._props.setdefault("js", {}).setdefault("constructors", {})
    shapes = JsHtmlStepper.JsShapes()
    for s in shapes.shapes:
      constructors[s] = "function %s(htmlObj, options, status){%s}" % (s, getattr(shapes, s)())
    return '<ul %s></ul>' % self.get_attrs(pyClassNames=self.style.get_classes())




from epyk.core.css import Classes
from epyk.core.css import Defaults_css
from epyk.core.css import Properties
from epyk.core.css.styles.effects import Effects
from epyk.core.css.styles.attributes import Attrs # for the rtype in the documentation
from epyk.core.css.styles.attributes import Commons, Body, Empty
from epyk.core.css.styles.classes import CssStyle, CssStyleScrollbar
from epyk.core.py import OrderedSet


class ClassPage(object):
  def __init__(self, htmlObj):
    self.htmlObj, self._css_struct, self._css_class = htmlObj, None, None
    self.__webkitscrollbar, self.__webkitscrollbar_track, self.__webkitscrollbar_thumb = None, None, None
    self.classList, self.__cls_defined, self.__cls_catalog = {"main": OrderedSet(), 'other': OrderedSet()}, None, None
    self.classList['other'].add(self.scrollbar_webkit)
    self.classList['other'].add(self.scrollbar_webkit_thumb)
    self.classList['other'].add(self.scrollbar_webkit_track)

  @property
  def css(self):
    """
    Description:
    ------------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: Commons
    """
    if self._css_struct is None:
      self._css_struct = Body(self.htmlObj)
    return self._css_struct

  @property
  def scrollbar_webkit(self):
    if not self.__webkitscrollbar:
      self.__webkitscrollbar = CssStyleScrollbar.CssWebkitScrollbar(self.htmlObj._report)
    return  self.__webkitscrollbar

  @property
  def scrollbar_webkit_thumb(self):
    if not self.__webkitscrollbar_thumb:
      self.__webkitscrollbar_thumb = CssStyleScrollbar.CssWebkitScrollbarThumb(self.htmlObj._report)
    return  self.__webkitscrollbar_thumb

  @property
  def scrollbar_webkit_track(self):
    if not self.__webkitscrollbar_track:
      self.__webkitscrollbar_track = CssStyleScrollbar.CssWebkitScrollbarTrack(self.htmlObj._report)
    return  self.__webkitscrollbar_track

  @property
  def defaults(self):
    """
    Description:
    ------------
    The Default CSS Attributes in the framework
    """
    return Defaults_css

  @property
  def add_classes(self):
    """
    Description:
    ------------
    Property to get access to the catalog of CSS classes to be added to the HTML class tag component

    :rtype: Classes.Catalog
    """
    if self.__cls_catalog is None:
      self.__cls_catalog = Classes.Catalog(self.htmlObj._report, self.classList)
    return self.__cls_catalog._class_type('main')

  @property
  def define_classes(self):
    """
    Description:
    ------------
    Property to get access to the catalog of CSS classes to be loaded in the page
    Those classes will not be automatically added to any HTML tag and they need to be added manually

    :rtype: Classes.Catalog
    """
    if self.__cls_catalog is None:
      self.__cls_catalog = Classes.Catalog(self.htmlObj._report, self.classList)
    return self.__cls_catalog._class_type('other')

  def get_classes(self):
    """
    Description:
    ------------
    Returns the list of Internal and bespoke classes to be added to the class HTML table on the component
    """
    for css_cls in self.classList.values():
      for c in css_cls:
        if hasattr(c, 'get_ref'):
          self.htmlObj._report._css[c.get_ref()] = str(c)
    return self.classList

  def get_classes_css(self):
    """
    Description:
    ------------

    :return:
    """
    css_frgs = {}
    for css_cls in self.classList.values():
      for c in css_cls:
        if hasattr(c, 'get_ref'):
          css_frgs[c.get_ref()] = str(c)
    return css_frgs

  def custom_class(self, css_attrs, classname=None, selector=None):
    """
    Description:
    -----------
    This will create dynamic CSS class which will not be added to any component.
    The class definition can then be reused in mutiple components.

    The CSS style of the body can only be done using predefined classes or inline CSS

    Usage::

      rptObj.body.style.custom_class(css_attrs={"_attrs": {"fill": 'red'}}, classname='nvd3.nv-pie .nv-pie-title')

    Attributes:
    ----------
    :param css_attrs: Nested dictionary with the different attributes
    :param classname: Optional. String. The classname in the CSS definition
    :param selector: Optional. String. The class selector (if it is not a classname using . but a strict definition)
    """
    if classname is None:
      cls_def = {"classname": False, '_selector': selector}
    else:
      cls_def = {"classname": classname}
    cls_def.update(css_attrs)
    v_cls = type(classname, (CssStyle.Style, ), cls_def)
    self.classList['other'].add(v_cls(self.htmlObj._report))
    return cls_def


class ClassHtml(Properties.CssMixin):
  def __init__(self, htmlObj):
    self.htmlObj, self._css_struct, self._css_class = htmlObj, None, None
    self.classList, self.__cls_defined, self.__cls_catalog = {"main": OrderedSet(), 'other': OrderedSet()}, None, None
    self.__cls_effects, self.__css_virtual = None, {}
    self.classList['main'].add(self.css_class)

  @property
  def varName(self):
    """
    Description:
    ------------
    Unique identifier for the CSS object on the Javascript side
    """
    return "%s_css" % self.htmlObj.htmlId

  @property
  def css(self):
    """
    Description:
    ------------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: Commons
    """
    if self._css_struct is None:
      self._css_struct = Commons(self.htmlObj)
    return self._css_struct

  @property
  def css_class(self):
    """
    Description:
    ------------
    The internal class used to put a custom Style to this object.
    Only 1 CSS class can be added to an HTML object

    :rtype: Classes.CatalogDiv.CatalogDiv
    """
    if self._css_class is None:
      self._css_class = Classes.CatalogDiv.CatalogDiv(self.htmlObj._report, self.classList['main'], html_id=self.htmlObj.htmlId).no_border()
    return self._css_class

  @property
  def defaults(self):
    """
    Description:
    ------------
    The Default CSS Attributes in the framework
    """
    return Defaults_css

  @property
  def effects(self):
    """
    Description:
    ------------
    Add animation effect to the component based either on a bespoke definition or a predefined one.

    :rtype: Effects.Effects
    """
    if self.__cls_effects is None:
      self.__cls_effects = Effects.Effects(self.htmlObj._report, self.htmlObj)
    return self.__cls_effects

  @property
  def add_classes(self):
    """
    Description:
    ------------
    Property to get access to the catalog of CSS classes to be added to the HTML class tag component

    :rtype: Classes.Catalog
    """
    if self.__cls_catalog is None:
      self.__cls_catalog = Classes.Catalog(self.htmlObj._report, self.classList)
    return self.__cls_catalog._class_type('main')

  @property
  def define_classes(self):
    """
    Description:
    ------------
    Property to get access to the catalog of CSS classes to be loaded in the page
    Those classes will not be automatically added to any HTML tag and they need to be added manually

    :rtype: Classes.Catalog
    """
    if self.__cls_catalog is None:
      self.__cls_catalog = Classes.Catalog(self.htmlObj._report, self.classList)
    return self.__cls_catalog._class_type('other')

  def attr(self, key, name, dflt=None, suffix="temp"):
    """
    Description:
    ------------
    The attr() function returns the value of an attribute of the selected elements.

    Related Pages:

			https://www.w3schools.com/cssref/func_attr.asp

    Attributes:
    ----------
    :param key:
    :param name:
    :param suffix:
    """
    key_selector = "_%s" % suffix
    if not key_selector in self.__css_virtual:
      self.__css_virtual[key_selector] = {}
    if dflt is not None:
      self.__css_virtual[key_selector].update({key: "attr(%s, %s)" % (name, dflt)})
    else:
      self.__css_virtual[key_selector].update({key: "attr(%s)" % name})

  def attr_content(self, name):
    """
    Description:
    ------------
    Use of the attr function for the before content value.
    This is the unique valid use of this CSS function in most of the browser

    Related Pages:

			https://gomakethings.com/how-to-access-and-use-data-attributes-in-your-css/

    Attributes:
    ----------
    :param name:
    """
    self.attr("content", name, suffix='before')

  def hover(self, attrs):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param attrs:
    """
    self.selector("hover", attrs)

  def selector(self, suffix, attrs):
    """
    Description:
    ------------

    Attributes:
    ----------
    :param suffix:
    :param attrs:
    """
    key_selector = "_%s" % suffix
    if not key_selector in self.__css_virtual:
      self.__css_virtual[key_selector] = {}
    self.__css_virtual[key_selector].update(attrs)

  def add_custom_class(self, classname, css_attrs):
    """
    Description:
    -----------
    This will create dynamic CSS class which will then be added to the component

    Usage::

      text = rptObj.ui.text("toto")
    text.style.add_custom_class("css_class", {"_attrs": {"color": 'red'}})

    Attributes:
    ----------
    :param classname: String. The classname
    :param css_attrs: Nested dictionary with the different attributes
    """
    cls_def = {"classname": classname}
    cls_def.update(css_attrs)
    v_cls = type(classname, (CssStyle.Style, ), cls_def)
    self.classList['main'].add(v_cls(self.htmlObj._report))
    return cls_def

  def no_class(self):
    """
    Description:
    ------------
    Clear all the Style, Classes and CSS attrubites for the HTML component.
    Once this function is called it is possible to add new CSS attributes or classes using the different catalog

    :return: self to allow the chaining
    """
    self.classList['main'] = OrderedSet()
    self._css_class = None
    return self

  def clear_style(self):
    """
    Description:
    ------------
    Clear all the inline CSS styles defined for the component

    :return: self to allow the chaining
    """
    self.htmlObj.attr['css'] = {}
    self.css.attrs = self.htmlObj.attr['css']
    return self

  def clear(self):
    """
    Description:
    ------------
    Remove the predefined class and set the default one for the div components

    :return: self to allow the chaining
    """
    self.classList['main'] = OrderedSet()
    if Defaults_css.DEFAULT_STYLE == 'no_border':
      self._css_class = Classes.CatalogDiv.CatalogDiv(self.htmlObj._report, self.classList['main'], html_id=self.htmlObj.htmlId).no_border()
    else:
      self._css_class = Defaults_css.DEFAULT_STYLE
    self.htmlObj.attr['class'] = self.classList['main']
    return self

  def clear_all(self):
    """
    Description:
    ------------
    Clear all the Style, Classes and CSS attrubites for the HTML component.
    Once this function is called it is possible to add new CSS attributes or classes using the different catalog
    Set the default style to no marging and no padding

    :return: self to allow the chaining
    """
    self.clear_style()
    self.clear()
    return self

  def builder(self, name, js_frg):
    """
    Description:
    ------------
    Attach a Javascript Builder to a CSS style.
    It will be triggered only once for all the HTML components using this style

    Attributes:
    ----------
    :param name: String. The Javascript variable name
    :param js_frg: String. The Javascript framework corresponding to the Js builder
    """
    self.htmlObj._report._props.setdefault('js', {}).setdefault("builders_css", OrderedSet()).add("const %s = %s" % (name, js_frg))
    return self

  def get_classes(self):
    """
    Description:
    ------------
    Returns the list of Internal and bespoke classes to be added to the class HTML table on the component
    """
    if self.__css_virtual and not '_attrs' in self.__css_virtual:
      self.__css_virtual["_attrs"] = self.__css_virtual.get('_temp', {})
      self.__css_virtual["_attrs"].update(dict(self.css.attrs))
      self.__css_virtual['classname'] = "style_%s" % self.htmlObj.htmlId
      meta_cls = type('Style%s' % self.htmlObj.htmlId, (CssStyle.Style,), self.__css_virtual)
      self.css.attrs = {} # empty the css inline section
      self.classList['main'].add(meta_cls(self.htmlObj._report))
      self.htmlObj.attr['css'] = {}
    for css_cls in self.classList.values():
      for c in css_cls:
        if hasattr(c, 'get_ref'):
          self.htmlObj._report._css[c.get_ref()] = c
    return self.classList

  def get_classes_css(self):
    """
    Description:
    ------------

    :return:
    """
    css_frgs = {}
    for css_cls in self.classList.values():
      for c in css_cls:
        if hasattr(c, 'get_ref'):
          css_frgs[c.get_ref()] = str(c)
    return css_frgs


class ClassHtmlEmpty(ClassHtml):

  @property
  def css(self):
    """
    Description:
    ------------
    Property to the underlying CSS definition to be added to the style HTML tag of a component

    :rtype: Commons
    """
    if self._css_struct is None:
      self._css_struct = Empty(self.htmlObj)
    return self._css_struct




from epyk.core.html.graph import GraphPlotly

from epyk.core.js.packages import JsPlotly


class DataHeader(GraphPlotly.DataChart):

  @property
  def align(self):
    return self._attrs["align"]

  @align.setter
  def align(self, val):
    self._attrs["align"] = val

  @property
  def values(self):
    return self._attrs["values"]

  @values.setter
  def values(self, val):
    self._attrs["values"] = val

  @property
  def line(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: GraphPlotly.DataMarkersLine
    """
    return self.sub_data("line", GraphPlotly.DataMarkersLine)

  @property
  def font(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: GraphPlotly.DataFont
    """
    return self.sub_data("font", GraphPlotly.DataFont)

  @property
  def fill(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: GraphPlotly.DataFill
    """
    return self.sub_data("fill", GraphPlotly.DataFill)


class DataCells(GraphPlotly.DataChart):

  @property
  def align(self):
    return self._attrs["align"]

  @align.setter
  def align(self, val):
    self._attrs["align"] = val

  @property
  def values(self):
    return self._attrs["values"]

  @values.setter
  def values(self, val):
    self._attrs["values"] = val

  @property
  def line(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: GraphPlotly.DataMarkersLine
    """
    return self.sub_data("line", GraphPlotly.DataMarkersLine)

  @property
  def font(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: GraphPlotly.DataFont
    """
    return self.sub_data("font", GraphPlotly.DataFont)

  @property
  def fill(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: GraphPlotly.DataFill
    """
    return self.sub_data("fill", GraphPlotly.DataFill)


class DataDomain(GraphPlotly.DataChart):

  @property
  def x(self):
    return self._attrs["x"]

  @x.setter
  def x(self, val):
    self._attrs["x"] = val

  @property
  def y(self):
    return self._attrs["y"]

  @y.setter
  def y(self, val):
    self._attrs["y"] = val


class DataTable(GraphPlotly.DataChart):

  @property
  def domain(self):
    """

    :rtype: DataDomain
    """
    return self.sub_data("domain", DataDomain)

  def set_domain(self, x, y=None):
    self.domain.x = x
    self.domain.y = y or x

  @property
  def columnorder(self):
    return self._attrs["columnorder"]

  @columnorder.setter
  def columnorder(self, val):
    self._attrs["columnorder"] = val

  @property
  def columnwidth(self):
    return self._attrs["columnwidth"]

  @columnwidth.setter
  def columnwidth(self, val):
    self._attrs["columnwidth"] = val

  @property
  def header(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: DataHeader
    """
    return self.sub_data("header", DataHeader)

  @property
  def cells(self):
    """

    https://plot.ly/javascript/bubble-maps/

    :rtype: DataCells
    """
    return self.sub_data("cells", DataCells)


class Table(GraphPlotly.Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """

    :rtype: JsPlotly.Pie
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: GraphPlotly.Layout
    """
    if self._layout is None:
      self._layout = GraphPlotly.Layout(self._report)
    return self._layout

  @property
  def data(self):
    return self._traces[-1]

  def headers_color(self, colors):
    """
    Description:
    ------------
    Set the background color of the header


    :param colors:
    """
    self.data.header.fill.color = colors
    return self

  def headers_font_color(self, colors):
    """

    :param colors:
    """
    self.data.header.font.color = colors
    return self

  def columns_color(self, colors):
    """

    :param colors:
    """
    self.data.cells.fill.color = colors
    return self

  def columns_font_color(self, colors):
    """

    :param colors:
    """
    self.data.cells.font.color = colors
    return self

  def add_trace(self, data, type='table', mode=None):
    """

    :param data:
    :param type:
    :param mode:
    """
    if type != 'table':
      c_data = dict(data)
      if type is not None:
        c_data['type'] = type
      if mode is not None:
        c_data['mode'] = mode
      trace = GraphPlotly.DataChart(self._report, attrs=c_data)
    else:
      c_data = {}
      if type is not None:
        c_data['type'] = type
      if mode is not None:
        c_data['mode'] = mode
      trace = DataTable(self._report, attrs=c_data)
      trace.cells.values = data
    self._traces.append(trace)
    return self




from epyk.core.html import Html
from epyk.core.html.options import OptPlotly

from epyk.core.js import JsUtils
from epyk.core.js.primitives import JsObject

from epyk.core.data import DataClass

from epyk.core.js.packages import JsPlotly
from epyk.core.js.packages import JsD3


class Chart(Html.Html):
  name, category, callFnc = 'Plotly', 'Charts', 'plotly'

  def __init__(self,  report, width, height, title, options, htmlCode, profile):
    self.seriesProperties, self.__chartJsEvents, self.height = {'static': {}, 'dynamic': {}}, {}, height[0]
    super(Chart, self).__init__(report, [], code=htmlCode, css_attrs={"width": width, "height": height}, profile=profile)
    self._d3, self._attrs, self._traces, self._layout, self._options = None, None, [], None, None
    self._options_init = options

  @property
  def chartId(self):
    """
    Return the Javascript variable of the chart
    """
    return "%s_obj" % self.htmlId

  @property
  def data(self):
    """

    :rtype: JsChartJs.DataSetPie
    :return:
    """
    if not self._traces:
      self.add_trace([])
    return self._traces[-1]

  @property
  def options(self):
    """

    :rtype: OptPlotly.OptionConfig
    :return:
    """
    if self._options is None:
      self._options = OptPlotly.OptionConfig(self._report, attrs=self._options_init)
    return self._options

  def traces(self, i=None):
    """

    :rtype: JsChartJs.DataSetPie
    """
    if i is None:
      return self._traces[-1]

    return self._traces[i]

  @property
  def layout(self):
    """

    :rtype: Layout
    :return:
    """
    if self._layout is None:
      self._layout = Layout(self._report)
    return self._layout

  @property
  def d3(self):
    """

    :rtype: JsD3.D3Select
    :return:
    """
    if self._d3 is None:
      self._d3 = JsD3.D3Select(self._report, selector="d3.select('#%s')" % self.htmlId, setVar=False)
    return self._d3

  def add_trace(self, data, type=None, mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataChart(self._report, attrs=c_data))
    return self

  def build(self, data=None, options=None, profile=False):
    str_traces = []
    for t in self._traces:
      str_traces.append("{%s}" % ", ".join(["%s: %s" % (k, JsUtils.jsConvertData(v, None)) for k, v in t.attrs()]))
    obj_datasets = JsObject.JsObject.get("[%s]" % ", ".join(str_traces))
    return JsUtils.jsConvertFncs([JsPlotly.JsPlotly(src=self._report).newPlot(self.htmlId, obj_datasets, self.layout, self.options)], toStr=True)

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<div %s></div>' % self.get_attrs(pyClassNames=self.style.get_classes())


class Line(Chart):
  __reqJs = ['plotly.js']

  @property
  def dom(self):
    """
    :rtype: JsPlotly.Line
    """
    if self._dom is None:
      self._dom = JsPlotly.Line(self, varName=self.chartId, report=self._report)
    return self._dom

  def add_trace(self, data, type=None, mode='lines+markers'):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataXY(self._report, attrs=c_data))
    return self


class Bar(Chart):
  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Bar(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    if self._layout is None:
      self._layout = LayoutBar(self._report)
    return self._layout

  def add_trace(self, data, type='bar', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataXY(self._report, attrs=c_data))
    return self


class DataFill(DataClass):

  @property
  def color(self):
    return self._attrs["color"]

  @color.setter
  def color(self, val):
    self._attrs["color"] = val


class LayoutAnnotation(DataClass):

  @property
  def xanchor(self):
    return self._attrs["xanchor"]

  @xanchor.setter
  def xanchor(self, val):
    self._attrs["xanchor"] = val

  @property
  def ax(self):
    return self._attrs["ax"]

  @ax.setter
  def ax(self, val):
    self._attrs["ax"] = val

  @property
  def ay(self):
    return self._attrs["ay"]

  @ay.setter
  def ay(self, val):
    self._attrs["ay"] = val

  @property
  def text(self):
    return self._attrs["text"]

  @text.setter
  def text(self, val):
    self._attrs["text"] = val

  @property
  def showarrow(self):
    return self._attrs["showarrow"]

  @showarrow.setter
  def showarrow(self, val):
    self._attrs["showarrow"] = val

  @property
  def x(self):
    return self._attrs["x"]

  @x.setter
  def x(self, val):
    self._attrs["x"] = val

  @property
  def y(self):
    return self._attrs["y"]

  @y.setter
  def y(self, val):
    self._attrs["y"] = val

  @property
  def xref(self):
    return self._attrs["xref"]

  @xref.setter
  def xref(self, val):
    self._attrs["xref"] = val

  @property
  def yref(self):
    return self._attrs["yref"]

  @yref.setter
  def yref(self, val):
    self._attrs["yref"] = val

  @property
  def font(self):
    return self.sub_data("font", DataFont)


class LayoutShape(DataClass):

  def add_path(self, points, color=None):
    """

    https://plot.ly/javascript/shapes/

    :param points:
    """
    self._attrs.update({'type': 'path', 'path': points})
    self.fillcolor = color or self._report.theme.warning[0]
    return self

  def add_line(self, x, y, x1, y1, opacity=0.2, color=None):
    """

    .layout.shapes.add_line(-100, 10, -50, -10, color="red")

    :param x:
    :param y:
    :param x1:
    :param y1:
    :param opacity:
    :param color:
    """
    self._attrs.update({'type': 'line', 'xref': 'x', 'yref': 'y', 'x0': x, 'y0': y, 'x1': x1, 'y1': y1})
    self.line.color = color or self._report.theme.warning[0]
    self.line.dash = 'dot'
    self.opacity = opacity
    return self

  def add_circle(self, x, y, x1, y1, opacity=0.2, color=None):
    """

    :param x:
    :param y:
    :param x1:
    :param y1:
    :param opacity:
    :param color:
    """
    self._attrs.update({'type': 'circle', 'xref': 'x', 'yref': 'y', 'x0': x, 'y0': y, 'x1': x1, 'y1': y1})
    self.fillcolor = color or self._report.theme.warning[0]
    self.line.width = 0
    self.opacity = opacity
    return self

  def add_rect(self, x, y, x1, y1, opacity=0.2, color=None):
    """

    :param x:
    :param y:
    :param x1:
    :param y1:
    """
    self._attrs.update({'type': 'rect', 'xref': 'x', 'yref': 'paper', 'x0': x, 'y0': y, 'x1': x1, 'y1': y1})
    self.fillcolor = color or self._report.theme.warning[0]
    self.line.width = 0
    self.opacity = opacity
    return self

  @property
  def type(self):
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val

  @property
  def xref(self):
    return self._attrs["xref"]

  @xref.setter
  def xref(self, val):
    self._attrs["xref"] = val

  @property
  def yref(self):
    return self._attrs["yref"]

  @yref.setter
  def yref(self, val):
    self._attrs["yref"] = val

  @property
  def x0(self):
    return self._attrs["x0"]

  @x0.setter
  def x0(self, val):
    self._attrs["x0"] = val

  @property
  def y0(self):
    return self._attrs["y0"]

  @y0.setter
  def y0(self, val):
    self._attrs["y0"] = val

  @property
  def x1(self):
    return self._attrs["x1"]

  @x1.setter
  def x1(self, val):
    self._attrs["x1"] = val

  @property
  def y1(self):
    return self._attrs["y1"]

  @y1.setter
  def y1(self, val):
    self._attrs["y1"] = val

  @property
  def fillcolor(self):
    return self._attrs["fillcolor"]

  @fillcolor.setter
  def fillcolor(self, val):
    self._attrs["fillcolor"] = val

  @property
  def opacity(self):
    return self._attrs["opacity"]

  @opacity.setter
  def opacity(self, val):
    self._attrs["opacity"] = val

  @property
  def line(self):
    """

    :rtype: DataMarkersLine
    :return:
    """
    return self.sub_data("line", DataMarkersLine)


class LayoutFont(DataClass):

  @property
  def color(self):
    """

    :param color: color or array of colors
    """
    return self._attrs["color"]

  @color.setter
  def color(self, val):
    self._attrs["color"] = val

  @property
  def family(self):
    """
    HTML font family - the typeface that will be applied by the web browser.

    https://plot.ly/javascript/reference/#pie-outsidetextfont-family

    :param font-family:  string or array of strings
    """
    return self._attrs["family"]

  @family.setter
  def family(self, val):
    self._attrs["family"] = val

  @property
  def size(self):
    """

    https://plot.ly/javascript/reference/#pie-outsidetextfont-family

    :return: number or array of numbers greater than or equal to 1
    """
    return self._attrs["size"]

  @size.setter
  def size(self, val):
    self._attrs["size"] = val


class LayoutGrid(DataClass):

  @property
  def rows(self):
    return self._attrs["rows"]

  @rows.setter
  def rows(self, val):
    self._attrs["rows"] = val

  @property
  def columns(self):
    return self._attrs["columns"]

  @columns.setter
  def columns(self, val):
    self._attrs["columns"] = val

  @property
  def pattern(self):
    return self._attrs["pattern"]

  @pattern.setter
  def pattern(self, val):
    self._attrs["pattern"] = val


class LayoutButtons(DataClass):

  @property
  def count(self):
    return self._attrs["count"]

  @count.setter
  def count(self, val):
    self._attrs["count"] = val

  @property
  def label(self):
    return self._attrs["label"]

  @label.setter
  def label(self, val):
    self._attrs["label"] = val

  @property
  def step(self):
    return self._attrs["step"]

  @step.setter
  def step(self, val):
    self._attrs["step"] = val

  @property
  def stepmode(self):
    return self._attrs["stepmode"]

  @stepmode.setter
  def stepmode(self, val):
    self._attrs["stepmode"] = val


class LayoutRangeSelector(DataClass):

  @property
  def buttons(self):
    """

    https://plot.ly/javascript/time-series/
    https://plot.ly/javascript/range-slider/

    :rtype: LayoutButtons
    """
    return self.sub_data_enum("buttons", LayoutButtons)

  def month(self, n):
    but = self.buttons
    but.step = 'month'
    but.stepmode = 'backward'
    but.count = n
    but.label = "%sm" % n
    return self

  def year(self, n=0):
    but = self.buttons
    but.step = 'year'
    but.count = n
    but.stepmode = 'todate' if n == 0 else 'backward'
    but.label = 'YTD' if n == 0 else "%sy" % n
    return self

  def all(self):
    but = self.buttons
    but.step = 'all'
    return self


class LayoutRangeSlider(DataClass):

  @property
  def range(self):
    return self._attrs["range"]

  @range.setter
  def range(self, val):
    self._attrs["range"] = val


class LayoutAxis(DataClass):

  @property
  def title(self):
    return self._attrs["title"]

  @title.setter
  def title(self, val):
    self._attrs["title"] = val

  @property
  def showbackground(self):
    return self._attrs["showbackground"]

  @showbackground.setter
  def showbackground(self, val):
    self._attrs["showbackground"] = val

  @property
  def backgroundcolor(self):
    return self._attrs["backgroundcolor"]

  @backgroundcolor.setter
  def backgroundcolor(self, val):
    self._attrs["backgroundcolor"] = val

  @property
  def tickangle(self):
    return self._attrs["tickangle"]

  @tickangle.setter
  def tickangle(self, val):
    self._attrs["tickangle"] = val

  @property
  def showticklabels(self):
    return self._attrs["showticklabels"]

  @showticklabels.setter
  def showticklabels(self, val):
    self._attrs["showticklabels"] = val

  @property
  def titlefont(self):
    """

    :rtype: LayoutFont

    :return:
    """
    return self.sub_data("titlefont", LayoutFont)

  @property
  def tickfont(self):
    """

    :rtype: LayoutFont

    :return:
    """
    return self.sub_data("tickfont", LayoutFont)

  def set_color(self, color):
    """

    :param color:
    """
    self.titlefont.color = color
    self.tickfont.color = color
    return self

  @property
  def overlaying(self):
    return self._attrs["overlaying"]

  @overlaying.setter
  def overlaying(self, val):
    self._attrs["overlaying"] = val

  @property
  def side(self):
    return self._attrs["side"]

  @side.setter
  def side(self, val):
    self._attrs["side"] = val

  @property
  def anchor(self):
    return self._attrs["anchor"]

  @anchor.setter
  def anchor(self, val):
    self._attrs["anchor"] = val

  @property
  def domain(self):
    return self._attrs["domain"]

  @domain.setter
  def domain(self, val):
    self._attrs["domain"] = val

  @property
  def dtick(self):
    return self._attrs["dtick"]

  @dtick.setter
  def dtick(self, val):
    self._attrs["dtick"] = val

  @property
  def autorange(self):
    return self._attrs["autorange"]

  @autorange.setter
  def autorange(self, val):
    self._attrs["autorange"] = val

  @property
  def position(self):
    return self._attrs["position"]

  @position.setter
  def position(self, val):
    self._attrs["position"] = val

  @property
  def paper_bgcolor(self):
    return self._attrs["paper_bgcolor"]

  @paper_bgcolor.setter
  def paper_bgcolor(self, val):
    self._attrs["paper_bgcolor"] = val

  @property
  def plot_bgcolor(self):
    return self._attrs["plot_bgcolor"]

  @plot_bgcolor.setter
  def plot_bgcolor(self, val):
    self._attrs["plot_bgcolor"] = val

  @property
  def range(self):
    return self._attrs["range"]

  @range.setter
  def range(self, val):
    self._attrs["range"] = val

  @property
  def gridcolor(self):
    return self._attrs["gridcolor"]

  @gridcolor.setter
  def gridcolor(self, val):
    self._attrs["gridcolor"] = val

  @property
  def gridwidth(self):
    return self._attrs["gridwidth"]

  @gridwidth.setter
  def gridwidth(self, val):
    self._attrs["gridwidth"] = val

  @property
  def type(self):
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val

  @property
  def zeroline(self):
    return self._attrs["zeroline"]

  @zeroline.setter
  def zeroline(self, val):
    self._attrs["zeroline"] = val

  @property
  def showline(self):
    return self._attrs["showline"]

  @showline.setter
  def showline(self, val):
    self._attrs["showline"] = val

  @property
  def showgrid(self):
    return self._attrs["showgrid"]

  @showgrid.setter
  def showgrid(self, val):
    self._attrs["showgrid"] = val

  @property
  def showticklabels(self):
    return self._attrs["showticklabels"]

  @showticklabels.setter
  def showticklabels(self, val):
    self._attrs["showticklabels"] = val

  @property
  def rangeselector(self):
    """

    https://plot.ly/javascript/time-series/

    :rtype: LayoutRangeSelector
    """
    return self.sub_data("rangeselector", LayoutRangeSelector)

  @property
  def rangeslider(self):
    """

    https://plot.ly/javascript/time-series/

    :rtype: LayoutRangeSlider
    """
    return self.sub_data("rangeslider", LayoutRangeSlider)


class LayoutMargin(DataClass):

  @property
  def l(self):
    return self._attrs["l"]

  @l.setter
  def l(self, val):
    self._attrs["l"] = val

  @property
  def r(self):
    return self._attrs["r"]

  @r.setter
  def r(self, val):
    self._attrs["r"] = val

  @property
  def b(self):
    return self._attrs["b"]

  @b.setter
  def b(self, val):
    self._attrs["b"] = val

  @property
  def t(self):
    return self._attrs["t"]

  @t.setter
  def t(self, val):
    self._attrs["t"] = val


class LayoutEye(DataClass):

  @property
  def x(self):
    return self._attrs["x"]

  @x.setter
  def x(self, val):
    self._attrs["x"] = val

  @property
  def y(self):
    return self._attrs["y"]

  @y.setter
  def y(self, val):
    self._attrs["y"] = val

  @property
  def z(self):
    return self._attrs["z"]

  @z.setter
  def z(self, val):
    self._attrs["z"] = val


class LayoutCamera(DataClass):

  @property
  def eye(self):
    """

    :rtype: LayoutEye

    :return:
    """
    return self.sub_data("eye", LayoutEye)


class LayoutScene(DataClass):

  @property
  def camera(self):
    """

    :rtype: LayoutCamera

    :return:
    """
    return self.sub_data("scene", LayoutCamera)

  @property
  def xaxis(self):
    """

    :rtype: LayoutAxis

    :return:
    """
    return self.sub_data("xaxis", LayoutAxis)

  @property
  def yaxis(self):
    """

    :rtype: LayoutAxis

    :return:
    """
    return self.sub_data("yaxis", LayoutAxis)

  @property
  def zaxis(self):
    """

    :rtype: LayoutAxis

    :return:
    """
    return self.sub_data("zaxis", LayoutAxis)


class LayoutLegend(DataClass):

  @property
  def x(self):
    return self._attrs["x"]

  @x.setter
  def x(self, val):
    self._attrs["x"] = val

  @property
  def bgcolor(self):
    return self._attrs["bgcolor"]

  @bgcolor.setter
  def bgcolor(self, val):
    self._attrs["bgcolor"] = val

  @property
  def bordercolor(self):
    return self._attrs["bordercolor"]

  @bordercolor.setter
  def bordercolor(self, val):
    self._attrs["bordercolor"] = val

  @property
  def borderwidth(self):
    return self._attrs["borderwidth"]

  @borderwidth.setter
  def borderwidth(self, val):
    self._attrs["borderwidth"] = val

  @property
  def traceorder(self):
    return self._attrs["traceorder"]

  @traceorder.setter
  def traceorder(self, val):
    self._attrs["traceorder"] = val

  @property
  def orientation(self):
    return self._attrs["orientation"]

  @orientation.setter
  def orientation(self, val):
    self._attrs["orientation"] = val

  @property
  def y(self):
    return self._attrs["y"]

  @y.setter
  def y(self, val):
    self._attrs["y"] = val

  @property
  def xanchor(self):
    return self._attrs["xanchor"]

  @xanchor.setter
  def xanchor(self, val):
    self._attrs["xanchor"] = val

  @property
  def font(self):
    """

    :rtype: LayoutFont
    """
    return self.sub_data("font", LayoutFont)


class Layout(DataClass):

  @property
  def title(self):
    return self._attrs["title"]

  @title.setter
  def title(self, val):
    self._attrs["title"] = val

  @property
  def paper_bgcolor(self):
    return self._attrs["paper_bgcolor"]

  @paper_bgcolor.setter
  def paper_bgcolor(self, val):
    self._attrs["paper_bgcolor"] = val

  @property
  def plot_bgcolor(self):
    return self._attrs["plot_bgcolor"]

  @plot_bgcolor.setter
  def plot_bgcolor(self, val):
    self._attrs["plot_bgcolor"] = val

  @property
  def showlegend(self):
    return self._attrs["showlegend"]

  @showlegend.setter
  def showlegend(self, val):
    self._attrs["showlegend"] = val

  @property
  def height(self):
    return self._attrs["height"]

  @height.setter
  def height(self, val):
    self._attrs["height"] = val

  @property
  def width(self):
    return self._attrs["width"]

  @width.setter
  def width(self, val):
    self._attrs["width"] = val

  @property
  def scene(self):
    """

    :rtype: LayoutScene
    """
    return self.sub_data("scene", LayoutScene)

  @property
  def legend(self):
    """

    https://plot.ly/javascript/legend/

    :rtype: LayoutLegend
    """
    return self.sub_data("legend", LayoutLegend)

  @property
  def xaxis(self):
    """

    https://plot.ly/javascript/time-series/

    :rtype: LayoutAxis
    """
    return self.sub_data("xaxis", LayoutAxis)

  @property
  def xaxis2(self):
    """

    https://plot.ly/javascript/time-series/

    :rtype: LayoutAxis
    """
    return self.sub_data("xaxis2", LayoutAxis)

  @property
  def grid(self):
    """

    https://plot.ly/javascript/subplots/

    :rtype: LayoutGrid
    """
    return self.sub_data("grid", LayoutGrid)

  @property
  def yaxis(self):
    """

    https://plot.ly/javascript/time-series/

    :rtype: LayoutAxis
    """
    return self.sub_data("yaxis", LayoutAxis)

  @property
  def yaxis2(self):
    """

    https://plot.ly/javascript/time-series/

    :rtype: LayoutAxis
    """
    return self.sub_data("yaxis2", LayoutAxis)

  @property
  def margin(self):
    """

    https://plot.ly/javascript/3d-surface-plots/

    :rtype: LayoutMargin
    """
    return self.sub_data("margin", LayoutMargin)

  def sub_plot(self, columns, rows=1, pattern='independent'):
    self.grid.rows = rows
    self.grid.columns = columns
    self.grid.pattern = pattern
    return self

  def inset_trace(self, x_domain, x, y=None, y_domain=None):
    """

    https://plot.ly/javascript/insets/

    :param x_domain:
    :param x:
    :param y:
    :param y_domain:
    """
    y = y or x
    y_domain = y_domain or x_domain
    x_axis = self.sub_data('xaxis%s' % x, LayoutAxis)
    x_axis.domain = x_domain
    x_axis.anchor = "y%s" % y
    y_axis = self.sub_data('yaxis%s' % y, LayoutAxis)
    y_axis.domain = y_domain
    y_axis.anchor = "x%s" % x
    return self

  def no_background(self):
    """

    https://community.plot.ly/t/you-can-remove-the-white-background-of-the-graphics-background/933

    :return:
    """
    self.paper_bgcolor = 'rgba(0,0,0,0)'
    self.plot_bgcolor = 'rgba(0,0,0,0)'
    return self

  def no_grid(self):
    """
    Remove the vertical and horizontal sub axis from the chart display.
    Keep the zeroline axis

    :return: The attribute object to allow the chaining
    """
    self.xaxis.showgrid = False
    self.xaxis.showline = False
    self.xaxis.showticklabels = False
    self.yaxis.showgrid = False
    self.yaxis.showline = False
    self.yaxis.showticklabels = False
    return self

  def grid_colors(self, x_color, y_color=None):
    """

    :param x_color:
    :param y_color:

    :return:
    """
    self.xaxis.gridcolor = x_color
    self.yaxis.gridcolor = y_color or x_color
    return self

  def axis_colors(self, x_color, y_color=None):
    """

    :param x_color:
    :param y_color:

    :return:
    """
    self.xaxis.set_color(x_color)
    self.yaxis.set_color(y_color or x_color)
    return self

  @property
  def shapes(self):
    """

    https://plot.ly/javascript/shapes/

    :rtype: LayoutShape
    """
    return self.sub_data_enum("shapes", LayoutShape)

  @property
  def annotations(self):
    """

    https://plot.ly/javascript/shapes/

    :rtype: LayoutAnnotation
    """
    return self.sub_data_enum("annotations", LayoutAnnotation)


class Layout3D(Layout):

  @property
  def scene(self):
    """

    :rtype: LayoutScene
    """
    return self.sub_data("scene", LayoutScene)

  def grid_colors(self, x_color, y_color=None, z_color=None):
    """

    :param x_color:
    :param y_color:
    :param z_color:
    """
    self.scene.xaxis.gridcolor = x_color
    self.scene.xaxis.zerolinecolor = x_color
    self.scene.yaxis.gridcolor = y_color
    self.scene.yaxis.zerolinecolor = y_color
    self.scene.zaxis.gridcolor = z_color
    self.scene.zaxis.zerolinecolor = z_color
    return self

  def axis_colors(self, x_color, y_color=None, z_color=None):
    """

    :param x_color:
    :param y_color:

    :return:
    """
    self.scene.xaxis.set_color(x_color)
    self.scene.yaxis.set_color(y_color or x_color)
    self.scene.zaxis.set_color(z_color or x_color)
    return self


class LayoutBar(Layout):

  @property
  def barmode(self):
    return self._attrs["barmode"]

  @barmode.setter
  def barmode(self, val):
    self._attrs["barmode"] = val


class LayoutBox(Layout):
  @property
  def boxmode(self):
    return self._attrs["boxmode"]

  @boxmode.setter
  def boxmode(self, val):
    self._attrs["boxmode"] = val


class DataFont(DataClass):

  @property
  def family(self):
    return self._attrs["family"]

  @family.setter
  def family(self, val):
    self._attrs["family"] = val

  @property
  def size(self):
    return self._attrs["size"]

  @size.setter
  def size(self, val):
    self._attrs["size"] = val

  @property
  def color(self):
    return self._attrs["color"]

  @color.setter
  def color(self, val):
    self._attrs["color"] = val


class DataMarkersLine(DataClass):

  @property
  def width(self):
    return self._attrs["width"]

  @width.setter
  def width(self, val):
    self._attrs["width"] = val

  @property
  def color(self):
    return self._attrs["color"]

  @color.setter
  def color(self, val):
    self._attrs["color"] = val

  @property
  def dash(self):
    return self._attrs["dash"]

  @dash.setter
  def dash(self, val):
    self._attrs["dash"] = val

  @property
  def outliercolor(self):
    return self._attrs["outliercolor"]

  @outliercolor.setter
  def outliercolor(self, val):
    self._attrs["outliercolor"] = val

  @property
  def outlierwidth(self):
    return self._attrs["outlierwidth"]

  @outlierwidth.setter
  def outlierwidth(self, val):
    self._attrs["outlierwidth"] = val


class DataMarkers(DataClass):

  @property
  def size(self):
    return self._attrs["size"]

  @size.setter
  def size(self, val):
    self._attrs["size"] = val

  @property
  def symbol(self):
    return self._attrs["symbol"]

  @symbol.setter
  def symbol(self, val):
    self._attrs["symbol"] = val

  @property
  def sizemode(self):
    return self._attrs["sizemode"]

  @sizemode.setter
  def sizemode(self, val):
    self._attrs["sizemode"] = val

  @property
  def color(self):
    return self._attrs["color"]

  @color.setter
  def color(self, val):
    self._attrs["color"] = val

  @property
  def opacity(self):
    return self._attrs["opacity"]

  @opacity.setter
  def opacity(self, val):
    self._attrs["opacity"] = val

  @property
  def width(self):
    return self._attrs["width"]

  @width.setter
  def width(self, val):
    self._attrs["width"] = val

  @property
  def outliercolor(self):
    """

    https://plot.ly/javascript/box-plots/
    """
    return self._attrs["outliercolor"]

  @outliercolor.setter
  def outliercolor(self, val):
    self._attrs["outliercolor"] = val

  @property
  def line(self):
    """

    https://plot.ly/javascript/webgl-vs-svg/

    :rtype: DataMarkersLine
    """
    return self.sub_data("line", DataMarkersLine)


class DataChart(DataClass):

  @property
  def automargin(self):
    return self._attrs["automargin"]

  @automargin.setter
  def automargin(self, val):
    self._attrs["automargin"] = val

  @property
  def hole(self):
    return self._attrs["hole"]

  @hole.setter
  def hole(self, val):
    self._attrs["hole"] = val

  @property
  def opacity(self):
    return self._attrs["opacity"]

  @opacity.setter
  def opacity(self, val):
    self._attrs["opacity"] = val

  @property
  def name(self):
    return self._attrs["name"]

  @name.setter
  def name(self, val):
    self._attrs["name"] = val

  @property
  def mode(self):
    return self._attrs["mode"]

  @mode.setter
  def mode(self, val):
    self._attrs["mode"] = val

  @property
  def fill(self):
    return self._attrs["fill"]

  @fill.setter
  def fill(self, val):
    self._attrs["fill"] = val

  @property
  def fillcolor(self):
    return self._attrs["fillcolor"]

  @fillcolor.setter
  def fillcolor(self, val):
    self._attrs["fillcolor"] = val

  @property
  def orientation(self):
    return self._attrs["orientation"]

  @orientation.setter
  def orientation(self, val):
    self._attrs["orientation"] = val

  @property
  def type(self):
    return self._attrs["type"]

  @type.setter
  def type(self, val):
    self._attrs["type"] = val

  @property
  def showlegend(self):
    return self._attrs["showlegend"]

  @showlegend.setter
  def showlegend(self, val):
    self._attrs["showlegend"] = val

  @property
  def legendgroup(self):
    return self._attrs["legendgroup"]

  @legendgroup.setter
  def legendgroup(self, val):
    self._attrs["legendgroup"] = val

  @property
  def mode(self):
    return self._attrs["mode"]

  @mode.setter
  def mode(self, val):
    self._attrs["mode"] = val

  @property
  def xaxis(self):
    return self._attrs["xaxis"]

  @xaxis.setter
  def xaxis(self, val):
    self._attrs["xaxis"] = val

  @property
  def yaxis(self):
    return self._attrs["yaxis"]

  @yaxis.setter
  def yaxis(self, val):
    self._attrs["yaxis"] = val

  def axis_index(self, x, y=None):
    """

    :param x:
    :param y:
    """
    self.xaxis = "x%s" % x
    self.yaxis = "y%s" % (y or x)
    return self

  @property
  def marker(self):
    """

    https://plot.ly/javascript/bubble-charts/

    :rtype: DataMarkers
    """
    return self.sub_data("marker", DataMarkers)


class DataXY(DataChart):

  @property
  def x(self):
    return self._attrs["x"]

  @x.setter
  def x(self, val):
    self._attrs["x"] = val

  @property
  def y(self):
    return self._attrs["y"]

  @y.setter
  def y(self, val):
    self._attrs["y"] = val


class DataPie(DataChart):

  @property
  def hole(self):
    return self._attrs["hole"]

  @hole.setter
  def hole(self, val):
    self._attrs["hole"] = val

  @property
  def values(self):
    return self._attrs["values"]

  @values.setter
  def values(self, val):
    self._attrs["values"] = val

  @property
  def labels(self):
    return self._attrs["labels"]

  @labels.setter
  def labels(self, val):
    self._attrs["labels"] = val

  @property
  def textinfo(self):
    return self._attrs["textinfo"]

  @textinfo.setter
  def textinfo(self, val):
    self._attrs["textinfo"] = val

  @property
  def textposition(self):
    return self._attrs["textposition"]

  @textposition.setter
  def textposition(self, val):
    self._attrs["textposition"] = val

  @property
  def outsidetextfont(self):
    """
    Sets the font used for `textinfo` lying outside the sector.

    https://plot.ly/javascript/reference/#pie-outsidetextfont-family

    :rtype: DataFont
    """
    return self.sub_data("outsidetextfont", DataFont)

  @property
  def hoverinfo(self):
    return self._attrs["hoverinfo"]

  @hoverinfo.setter
  def hoverinfo(self, val):
    self._attrs["hoverinfo"] = val

  @property
  def text(self):
    return self._attrs["text"]

  @text.setter
  def text(self, val):
    self._attrs["text"] = val

  @property
  def textposition(self):
    return self._attrs["textposition"]

  @textposition.setter
  def textposition(self, val):
    self._attrs["textposition"] = val


class DataProject(DataChart):

  @property
  def z(self):
    return self._attrs["z"]

  @z.setter
  def z(self, val):
    self._attrs["z"] = val


class DataZ(DataChart):

  @property
  def show(self):
    return self._attrs["show"]

  @show.setter
  def show(self, val):
    self._attrs["show"] = val

  @property
  def usecolormap(self):
    return self._attrs["usecolormap"]

  @usecolormap.setter
  def usecolormap(self, val):
    self._attrs["usecolormap"] = val

  @property
  def highlightcolor(self):
    return self._attrs["highlightcolor"]

  @highlightcolor.setter
  def highlightcolor(self, val):
    self._attrs["highlightcolor"] = val

  @property
  def project(self):
    """

    https://plot.ly/javascript/3d-surface-plots/

    :rtype: DataProject
    """
    return self.sub_data("project", DataProject)


class DataContours(DataChart):

  @property
  def z(self):
    """

    https://plot.ly/javascript/3d-surface-plots/

    :rtype: DataZ
    """
    return self.sub_data("z", DataZ)


class DataLine(DataChart):

  @property
  def color(self):
    return self._attrs["color"]

  @color.setter
  def color(self, val):
    self._attrs["color"] = val


class DataMove(DataChart):

  @property
  def line(self):
    """

    :rtype: DataLine

    :return:
    """
    return self.sub_data("line", DataLine)


class DataSurface(DataChart):

  @property
  def showscale(self):
    return self._attrs["showscale"]

  @showscale.setter
  def showscale(self, val):
    self._attrs["showscale"] = val

  @property
  def contours(self):
    """

    https://plot.ly/javascript/3d-surface-plots/

    :rtype: DataContours
    """
    return self.sub_data("contours", DataContours)

  @property
  def line(self):
    """

    :rtype: DataLine

    :return:
    """
    return self.sub_data("line", DataLine)


class DataDelta(DataClass):

  @property
  def reference(self):
    return self._attrs["reference"]

  @reference.setter
  def reference(self, val):
    self._attrs["reference"] = val

  @property
  def relative(self):
    return self._attrs["relative"]

  @relative.setter
  def relative(self, val):
    self._attrs["relative"] = val

  @property
  def position(self):
    return self._attrs["position"]

  @position.setter
  def position(self, val):
    self._attrs["position"] = val

  @property
  def valueformat(self):
    return self._attrs["valueformat"]

  @valueformat.setter
  def valueformat(self, val):
    self._attrs["valueformat"] = val


class DataTitle(DataChart):

  @property
  def text(self):
    return self._attrs["text"]

  @text.setter
  def text(self, val):
    self._attrs["text"] = val


class DataNumber(DataChart):

  @property
  def prefix(self):
    return self._attrs["prefix"]

  @prefix.setter
  def prefix(self, val):
    self._attrs["prefix"] = val


class DataGauge(DataChart):

  @property
  def shape(self):
    return self._attrs["shape"]

  @shape.setter
  def shape(self, val):
    self._attrs["shape"] = val

  @property
  def axis(self):
    """

    https://plot.ly/javascript/indicator/

    :rtype: LayoutAxis
    """
    return self.sub_data("axis", LayoutAxis)


class DataIndicator(DataChart):

  @property
  def vmax(self):
    return self._attrs["vmax"]

  @vmax.setter
  def vmax(self, val):
    self._attrs["vmax"] = val

  def domain(self, x, y):
    """

    https://plot.ly/javascript/indicator/

    :param x:
    :param y:
    """
    self._attrs['domain'] = {"x": x, 'y': y}

  @property
  def title(self):
    """

    https://plot.ly/javascript/indicator/

    :rtype: DataTitle
    """
    return self.sub_data("title", DataTitle)

  @property
  def number(self):
    """

    https://plot.ly/javascript/indicator/

    :rtype: DataNumber
    """
    return self.sub_data("number", DataNumber)

  @property
  def gauge(self):
    """

    https://plot.ly/javascript/indicator/

    :rtype: DataGauge
    """
    if not 'gauge' in self.mode:
      self.mode = "%s+gauge" % self.mode
    return self.sub_data("gauge", DataGauge)

  def add_prefix(self, text):
    self.number.prefix = text
    return self

  def add_title(self, text):
    """

    delta.data.add_title("<b style='color:red'>test</b>")

    https://plot.ly/javascript/indicator/

    :param text:

    :return:
    """
    self.title.text = text
    return self

  @property
  def delta(self):
    """

    https://plot.ly/javascript/3d-surface-plots/

    :rtype: DataDelta
    """
    return self.sub_data("delta", DataDelta)


class DataBox(DataChart):

  @property
  def boxpoints(self):
    return self._attrs["boxpoints"]

  @boxpoints.setter
  def boxpoints(self, val):
    self._attrs["boxpoints"] = val

  @property
  def boxmean(self):
    return self._attrs["boxmean"]

  @boxmean.setter
  def boxmean(self, val):
    self._attrs["boxmean"] = val

  @property
  def jitter(self):
    return self._attrs["jitter"]

  @jitter.setter
  def jitter(self, val):
    self._attrs["jitter"] = val

  @property
  def whiskerwidth(self):
    return self._attrs["whiskerwidth"]

  @whiskerwidth.setter
  def whiskerwidth(self, val):
    self._attrs["whiskerwidth"] = val

  @property
  def pointpos(self):
    return self._attrs["pointpos"]

  @pointpos.setter
  def pointpos(self, val):
    self._attrs["pointpos"] = val


class DataCandle(DataChart):

  @property
  def close(self):
    return self._attrs["close"]

  @close.setter
  def close(self, val):
    self._attrs["close"] = val

  @property
  def high(self):
    return self._attrs["high"]

  @high.setter
  def high(self, val):
    self._attrs["high"] = val

  @property
  def low(self):
    return self._attrs["low"]

  @low.setter
  def low(self, val):
    self._attrs["low"] = val

  @property
  def open(self):
    return self._attrs["open"]

  @open.setter
  def open(self, val):
    self._attrs["open"] = val

  @property
  def increasing(self):
    """

    :rtype: DataMove

    :return:
    """
    return self.sub_data("increasing", DataMove)

  @property
  def decreasing(self):
    """

    :rtype: DataMove

    :return:
    """
    return self.sub_data("decreasing", DataMove)


class Pie(Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Pie
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def data(self):
    """

    :rtype: DataPie

    :return:
    """
    if not self._traces:
      self._traces.append(DataPie(self._report))
    return self._traces[-1]

  def add_trace(self, data, type='pie', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataPie(self._report, attrs=c_data))
    return self


class Surface(Chart):
  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Pie
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='surface', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataSurface(self._report, attrs=c_data))
    return self


class Scatter3D(Chart):
  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Pie
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    if self._layout is None:
      self._layout = Layout3D(self._report)
    return self._layout

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='scatter3d', mode="lines"):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataSurface(self._report, attrs=c_data))
    return self


class Mesh3d(Chart):
  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Pie
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def data(self):
    return self._traces[-1]

  def add_trace(self, data, type='mesh3d', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataSurface(self._report, attrs=c_data))
    return self


class Indicator(Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  def add_trace(self, data, type='indicator', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataIndicator(self._report, attrs=c_data))
    return self


class ScatterPolar(Chart):

  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  def add_trace(self, data, type='scatterpolar', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataChart(self._report, attrs=c_data))
    self.data.fill = 'toself'
    return self


class Box(Chart):
  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: LayoutBox

    :return:
    """
    if self._layout is None:
      self._layout = LayoutBox(self._report)
    return self._layout

  def add_trace(self, data, type='box', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataBox(self._report, attrs=c_data))
    return self


class CandleStick(Chart):
  __reqJs = ['plotly.js']

  @property
  def chart(self):
    """
    :rtype: JsPlotly.Bar
    """
    if self._chart is None:
      self._chart = JsPlotly.Pie(self._report, varName=self.chartId)
    return self._chart

  @property
  def layout(self):
    """

    :rtype: LayoutBox

    :return:
    """
    if self._layout is None:
      self._layout = LayoutBox(self._report)
    return self._layout

  def add_trace(self, data, type='candlestick', mode=None):
    c_data = dict(data)
    if type is not None:
      c_data['type'] = type
    if mode is not None:
      c_data['mode'] = mode
    self._traces.append(DataCandle(self._report, attrs=c_data))
    return self




from epyk.core.html import Defaults as Defaults_html
from epyk.core.css import Defaults as Defaults_css

from epyk.core.css.styles.classes import CssStyle


class CssInput(CssStyle.Style):
  """
  CSS Base style for the input components
  """
  _attrs = {'border': 'none', 'text-align': 'center', 'cursor': 'text', 'margin': 0, 'border-radius': '5px'}
  _focus = {'outline': 0}

  def customize(self):
    self.attrs.css({"background": self.rptObj.theme.colors[0], "color": self.rptObj.theme.greys[-1],
                    'font-family': Defaults_css.Font.family, 'min-width': '%spx' % Defaults_html.INPUTS_MIN_WIDTH,
                    'line-height': '%spx' % Defaults_html.LINE_HEIGHT,
                    'border': '1px solid %s' % self.rptObj.theme.colors[0],
                    'font-size': '%s%s' % (Defaults_css.Font.size, Defaults_css.Font.unit)})
    self.hover.css({'color': self.rptObj.theme.success[1]})


class CssInputRange(CssStyle.Style):
  """
  CSS Style for the input range component
  """
  _attrs = {'-webkit-appearance': 'none', 'appearance': 'none', 'outline': 'none', 'opacity': 0.7,
            '-webkit-transition': '.2s', 'transition': 'opacity .2s', 'cursor': 'pointer'}
  _hover = {'opacity': 1}
  _selectors = {'child': 'input'}

  def customize(self):
    self.attrs.css({"background": self.rptObj.theme.colors[0]})


class CssInputRangeThumb(CssStyle.Style):
  """
  CSS Style for the thumb of the input range component
  """
  _attrs = {'-webkit-appearance': 'none', 'appearance': 'none', 'cursor': 'pointer'}
  _webkit_slider_thumb = {'-webkit-appearance': 'none', 'appearance': 'none', 'cursor': 'pointer'}
  _selectors = {'child': 'input'}

  def customize(self):
    self.webkit_slider_thumb.css({"background": self.rptObj.theme.success[1],
                    'width': '%spx' % Defaults_html.INPUTS_RANGE_THUMB,
                    'height': '%spx' % Defaults_html.INPUTS_RANGE_THUMB})


class CssInputLabel(CssStyle.Style):
  """
  CSS Style for the label attached to an input component
  """
  _attrs = {'line-height': '1.5', 'margin-left': '10px'}
  _selectors = {'child': 'label'}

  def customize(self):
    self.attrs.css({'font-family': Defaults_css.Font.family,
                    'font-size': '%s%s' % (Defaults_css.Font.size, Defaults_css.Font.unit)})


class CssInputInteger(CssStyle.Style):
  """

  """
  cssId = {'reference': 'input[type=number]::-webkit-inner-spin-button, input[type=number]::-webkit-outer-spin-button'}
  _attrs = {"-webkit-appearance": 'none', "margin": 0}


class CssInputText(CssStyle.Style):
  """
  CSS Style for the input text component (within a field object)
  """
  _attrs = {'margin-left': '10px'}
  _selectors = {'child': 'input'}


class CssInputTextArea(CssStyle.Style):
  """
  CSS Style for the textarea component.
  """
  _attrs = {'resize': 'none', 'margin-bottom': '5px'}
  _focus = {'outline': 0}

  def customize(self):
    self.css({"background-color": self.rptObj.theme.colors[0], "color": self.rptObj.theme.greys[-1],
              'border': '1px solid %s' % self.rptObj.theme.colors[1]})
    self.hover.css({'color': self.rptObj.theme.greys[-1]})


class CssInputValid(CssStyle.Style):
  _valid = {'color': 'red'}
  _invalid = {'color': 'yellow', "background": "url(https://s3-us-west-2.amazonaws.com/s.cdpn.io/3/check.svg)",
            "background-size": "10px", "background-repeat": 'no-repeat', "background-position": "0"}


class CssUIActive(CssStyle.Style):
  classname = "ui-state-active"

  def customize(self):
    self.css({"border": "1px solid %s" % self.rptObj.theme.success[1], 'background-color': self.rptObj.theme.success[1]}, important=True)
    self.hover.css({"border": "1px solid %s" % self.rptObj.theme.success[1],
                    'background-color': self.rptObj.theme.success[1]})





from epyk.core.html.options import Options


class OptionsSelect(Options):

  @property
  def actionsBox(self):
    """
    Description:
    ------------
    When set to true, adds two buttons to the top of the dropdown menu (Select All & Deselect All).

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @actionsBox.setter
  def actionsBox(self, bool):
    self._config(bool)

  @property
  def container(self):
    """
    Description:
    ------------
    When set to a string, appends the select to a specific element or selector, e.g., container: 'body' | '.main-body'

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @container.setter
  def container(self, bool):
    self._config(bool)

  @property
  def deselectAllText(self):
    """
    Description:
    ------------
    The text on the button that deselects all options when actionsBox is enabled.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @deselectAllText.setter
  def deselectAllText(self, value):
    self._config(value)

  @property
  def dropdownAlignRight(self):
    """
    Description:
    ------------
    Align the menu to the right instead of the left. If set to 'auto', the menu will automatically align right if there isn't room for the menu's full width when aligned to the left.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @dropdownAlignRight.setter
  def dropdownAlignRight(self, bool):
    self._config(bool)

  @property
  def dropupAuto(self):
    """
    Description:
    ------------
    checks to see which has more room, above or below.
    If the dropup has enough room to fully open normally, but there is more room above, the dropup still opens normally.
    Otherwise, it becomes a dropup. If dropupAuto is set to false, dropups must be called manually.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @dropupAuto.setter
  def dropupAuto(self, bool):
    self._config(bool)

  @property
  def header(self):
    """
    Description:
    ------------
    adds a header to the top of the menu; includes a close button by default

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @header.setter
  def header(self, bool):
    self._config(bool)

  @property
  def hideDisabled(self):
    """
    Description:
    ------------
    removes disabled options and optgroups from the menu data-hide-disabled: true

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @hideDisabled.setter
  def hideDisabled(self, bool):
    self._config(bool)

  @property
  def iconBase(self):
    """
    Description:
    ------------
    Set the base to use a different icon font instead of Glyphicons.
    If changing iconBase, you might also want to change tickIcon, in case the new icon font uses a different naming scheme.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @iconBase.setter
  def iconBase(self, bool):
    self._config(bool)

  @property
  def liveSearch(self):
    """
    Description:
    ------------
    When set to true, adds a search box to the top of the selectpicker dropdown.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @liveSearch.setter
  def liveSearch(self, bool):
    self._config(bool)

  @property
  def liveSearchPlaceholder(self):
    """
    Description:
    ------------
    When set to a string, a placeholder attribute equal to the string will be added to the liveSearch input.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @liveSearchPlaceholder.setter
  def liveSearchPlaceholder(self, value):
    self._config(value)

  @property
  def liveSearchStyle(self):
    """
    Description:
    ------------
    When set to 'contains', searching will reveal options that contain the searched text.
    For example, searching for pl with return both Apple, Plum, and Plantain.
    When set to 'startsWith', searching for pl will return only Plum and Plantain.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @liveSearchStyle.setter
  def liveSearchStyle(self, value):
    self._config(value)

  @property
  def maxOptions(self):
    """
    Description:
    ------------
    When set to an integer and in a multi-select, the number of selected options cannot exceed the given value.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @maxOptions.setter
  def maxOptions(self, value):
    self._config(value)

  @property
  def maxOptionsText(self):
    """
    Description:
    ------------
    The text that is displayed when maxOptions is enabled and the maximum number of options for the given scenario have been selected.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @maxOptionsText.setter
  def maxOptionsText(self, value):
    self._config(value)

  @property
  def mobile(self):
    """
    Description:
    ------------
    When set to true, enables the device's native menu for select menus.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @mobile.setter
  def mobile(self, bool):
    self._config(bool)

  @property
  def noneSelectedText(self):
    """
    Description:
    ------------
    The text that is displayed when a multiple select has no selected options.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @noneSelectedText.setter
  def noneSelectedText(self, value):
    self._config(value)

  @property
  def noneResultsText(self):
    """
    Description:
    ------------
    The text displayed when a search doesn't return any results.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @noneResultsText.setter
  def noneResultsText(self, value):
    self._config(value)

  @property
  def selectAllText(self):
    """
    Description:
    ------------
    The text on the button that selects all options when actionsBox is enabled.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @selectAllText.setter
  def selectAllText(self, value):
    self._config(value)

  @property
  def selectOnTab(self):
    """
    Description:
    ------------
    When set to true, treats the tab character like the enter or space characters within the selectpicker dropdown.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @selectOnTab.setter
  def selectOnTab(self, bool):
    self._config(bool)

  @property
  def showContent(self):
    """
    Description:
    ------------
    When set to true, display custom HTML associated with selected option(s) in the button.
    When set to false, the option value will be displayed instead.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @showContent.setter
  def showContent(self, bool):
    self._config(bool)

  @property
  def showIcon(self):
    """
    Description:
    ------------
    When set to true, display icon(s) associated with selected option(s) in the button.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @showIcon.setter
  def showIcon(self, bool):
    self._config(bool)

  @property
  def showSubtext(self):
    """
    Description:
    ------------
    When set to true, display subtext associated with a selected option in the button.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @showSubtext.setter
  def showSubtext(self, bool):
    self._config(bool)

  @property
  def showTick(self):
    """
    Description:
    ------------
    Show checkmark on selected option (for items without multiple attribute).

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @showTick.setter
  def showTick(self, bool):
    self._config(bool)

  @property
  def size(self):
    """
    Description:
    ------------
    When set to 'auto', the menu always opens up to show as many items as the window will allow without being cut off.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @size.setter
  def size(self, value):
    self._config(value)

  @property
  def style(self):
    """
    Description:
    ------------
    When set to a string, add the value to the button's style.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @style.setter
  def style(self, value):
    self._config(value)

  @property
  def styleBase(self):
    """
    Description:
    ------------
    The default class applied to the button. When using the setStyle method, this class will always remain.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @styleBase.setter
  def styleBase(self, value):
    self._config(value)

  @property
  def tickIcon(self):
    """
    Description:
    ------------
    Set which icon to use to display as the "tick" next to selected options.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @tickIcon.setter
  def tickIcon(self, value):
    self._config(value)

  @property
  def title(self):
    """
    Description:
    ------------
    The default title for the selectpicker.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @title.setter
  def title(self, value):
    self._config(value)

  @property
  def virtualScroll(self):
    """
    Description:
    ------------
    If enabled, the items in the dropdown will be rendered using virtualization (i.e. only the items that are within the viewport will be rendered).
    This drastically improves performance for selects with a large number of options.
    Set to an integer to only use virtualization if the select has at least that number of options.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @virtualScroll.setter
  def virtualScroll(self, value):
    self._config(value)

  @property
  def width(self):
    """
    Description:
    ------------
    When set to auto, the width of the selectpicker is automatically adjusted to accommodate the widest option.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @width.setter
  def width(self, value):
    self._config(value)

  @property
  def windowPadding(self):
    """
    Description:
    ------------
    This is useful in cases where the window has areas that the dropdown menu should not cover - for instance a fixed header.
    When set to an integer, the same padding will be added to all sides.
    Alternatively, an array of integers can be used in the format [top, right, bottom, left].

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @windowPadding.setter
  def windowPadding(self, value):
    self._config(value)

  @property
  def sanitize(self):
    """
    Description:
    ------------
    Enable or disable the sanitization. If activated, 'data-content' on individual options will be sanitized.

    Related Pages:

			https://developer.snapappointments.com/bootstrap-select/options/
    """
    return self._config_get(False)

  @sanitize.setter
  def sanitize(self, bool):
    self._config(bool)



"""

"""

from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.tests import test_statics

from epyk.core.js.packages import JsChartJs
from epyk.core.Page import Report


# Create a blank Javascript Python object
rptObj = Report()

f = JsUtils.JsFile("DataExamplesChartJs", path=test_statics.OUTPUT_PATHS)

data = test_statics.get_data("flights.txt", n=10)
record = rptObj.js.data.records(data)
data_fnc = record.fnc.count(["origin", "destination"], ["distance", "delay"])

chart_config_line = record.to.chartJs.line(["distance", "delay"], "origin", attrs={"fill": False})
print(chart_config_line.toStr())

chart = JsChartJs.ChartJs("test", JsChartJs.ChartJsType("line", chart_config_line), rptObj, varName="chart_test")


f.writeJs([
  #jsObj.console.log(data_fnc),
  #rptObj.js.console.log(record.to.chartJs.line(["distance"], "origin")),
  rptObj.js.createElement(tagName="canvas", varName="test").attr("id", "test"),
  rptObj.js.body.appendChild(rptObj.js.objects.dom.get("test")),
  #rptObj.js.console.log(rptObj.js.objects.dom.get("test")),
  rptObj.js.console.log(chart_config_line),
  chart,
])

#f.writeJs([
  #jsObj.console.log(data_fnc),
#  rptObj.js.console.log(record.to.chartJs.line(["distance"], "origin")),
  #rptObj.js.console.log(chart)
#])

# Close the file and print the location of the launcher
print(f.close(rptObj))



"""
Module dedicated to perform the data transformation for the Vis charts

Related Pages:

		https://visjs.org/
"""

from epyk.core.js import JsUtils


class JsChartVisLinks(object):
  def __init__(self, data, js_src, data_schema=None, profile=False):
    self._js_src, self._data_schema, self.profile, self._data = js_src, data_schema, profile, data

  def __register_records_fnc(self, fnc_name, fnc_def, fnc_pmts=None, profile=False):
    """
    This function will attach to the report object only the javascript functions used during the report

    :param fnc_name: A String with the Javascript function name to be defined
    :param fnc_def: A String with the Javascript function content
    :param fnc_pmts: A list of parameters
    :param profile: A boolean flag to activate the framework profiling

    :return:
    """
    fnc_pmts = ["data"] + (fnc_pmts or [])
    if not fnc_name in self._js_src.get('js', {}).get('functions', {}):
      if profile:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      else:
        content = "var result = []; %s;return result" % JsUtils.cleanFncs(fnc_def)
      self._js_src.setdefault('js', {}).setdefault('functions', {})[fnc_name] = {'content': content, 'pmt': fnc_pmts}

  def timeline(self):
    pass

  def network(self):
    pass

  def two_dim(self, y_columns, x_axis, group=None, profile=False):
    """

    :param y_columns:
    :param x_axis:
    :param group:
    :param profile:

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsVis.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsVis.content, list(JsVis.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s, %s)" % (fnc_name, y_columns, x_axis, JsUtils.jsConvertData(group, None))

  def three_dim(self, y_columns, x_axis, z_axis, profile=False):
    """

    :param y_columns:
    :param x_axis:
    :param z_axis:
    :param profile:

    :return:
    """
    if not isinstance(y_columns, list):
      y_columns = [y_columns]
    fnc_name = JsVis.__name__
    x_axis = JsUtils.jsConvertData(x_axis, None)
    self.__register_records_fnc(fnc_name, JsVis3D.content, list(JsVis3D.pmts), profile | self.profile)
    self._data_schema['out'] = "%s(%%s, %s, %s, %s)" % (fnc_name, y_columns, x_axis, JsUtils.jsConvertData(z_axis, None))


# --------------------------------------------------------------------------------------------------------------------
#                               DATA TRANSFORMATION JAVASCRIPT DEFINITION
# --------------------------------------------------------------------------------------------------------------------
class JsVis(object):
  """
  :category: Vis Transformer
  :rubric: JS
  :type: Configuration
  :dsc:
    Transform a recordset to a data structure usable for a Vis 2D chart.
    As Vis is used for time line, the x axis should be a date in the format YYYY-MM-DD.
    Example of DataFrame:

      index x           y
      0     2018-01-01  10
      1     2018-01-02  20
      2     2018-01-03  30

  :example: aresObj.chart('bar', df, seriesNames=['y', 'y2'], xAxis='x', chartFamily='Vis')
  :example: aresObj.chart('line', df, seriesNames=['y', 'y2'], xAxis='x', chartFamily='Vis')
  """
  pmts = ("seriesNames", "xAxis", "group")
  content = '''
    var temp = {}; var labels = {}; var groups = {};
    data.forEach(function(rec){ 
      if(!(rec[xAxis] in temp)){temp[rec[xAxis]] = {}; groups[rec[xAxis]] = {}};
      seriesNames.forEach(function(name){
        labels[name] = true; if(rec[name] !== undefined){
          if(!(name in temp[rec[xAxis]])){
            temp[rec[xAxis]][name] = rec[name]; if(group !== undefined){groups[rec[xAxis]][name] = rec[group]}} 
          else{temp[rec[xAxis]][name] += rec[name]}}})});
    var labels = Object.keys(labels);
    for(var series in temp){
      labels.forEach(function(label, i){if(temp[series][label] !== undefined){ 
        var row = {x: series, y: temp[series][label], label: {content: temp[series][label].formatMoney(1, ',', '.')}};
        if(group !== undefined){row['group'] = groups[series][label]} else{row['group'] = i};
        result.push(row)}})}
    '''


class JsVis3D(object):
  """
  :category: Vis Transformer
  :rubric: JS
  :type: Configuration
  :dsc:


    for the surface the recordset should be
      data = [
        { x: 0, y: 0, z: 50 },
        { x: 0, y: 31.4, z: 50 },
        { x: 0, y: 62.8, z: 50 },
        { x: 31.4, y: 0, z: 79.37637628569459 },
        { x: 31.4, y: 31.4, z: 20.623660971554436 },
        { x: 31.4, y: 62.8, z: 26.179684010439026 }
      ]

    style: rec[zAxis]
  """
  chartTypes = ['3D', 'surface', 'bar3d', 'scatter3d', 'line3d', 'bubble']
  pmts = ("yAxis", "xAxis", "zAxis")
  content = '''
    data.forEach(function(rec, i){ 
      yAxis.forEach(function(s, j){
        result.push({x: parseFloat(rec[xAxis]), y: rec[s], z: rec[zAxis], name: s, style: rec[zAxis]})})})
    '''




from epyk.core.js.objects import JsNodeDom
from epyk.core.js.packages import JsTabulator


class JsHtmlTabulator(JsNodeDom.JsDoms):

  def __init__(self, htmlObj):
    super(JsHtmlTabulator, self).__init__(htmlObj)
    self._tabulator = None

  @property
  def tabulator(self):
    """

    :return:
    :rtype: JsTabulator.Tabulator
    """
    if self._tabulator is None:
      self._tabulator = JsTabulator.Tabulator(self._src)
    return self._tabulator



from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.tests import test_statics

# Create a blank Javascript Python object
jsObj = Js.JsBase()

# Define a dataSet using CrossFilter library
crossFilter = jsObj.data.crossfilter(data=[{"column": 200}], var_name="test")

# Write the results to a Javascript file with a Launcher
f = JsUtils.JsFile("CrossFilter", path=test_statics.OUTPUT_PATHS)

# Write the Javascript fragments to the file
f.writeJs([
  crossFilter.toStr(),
  crossFilter.dimension("column").filterAll().filterRange(100, 400).top(10).toStr(),
  jsObj.console.log(crossFilter.var)
])

# Close the file and print the location of the launcher
print(f.close(jsObj))


"""

"""

from epyk.core.css import Defaults_css
from epyk.core.css.styles.attributes import Attrs


class AttrInput(Attrs):
  def __init__(self, rptObj):
    super(AttrInput, self).__init__(rptObj)
    self.font_size = Defaults_css.font()
    self.font_family = Defaults_css.Font.family
    self.box_sizing = 'border-box'




from epyk.core.js import JsUtils

from epyk.core.js.packages import JsPackage
from epyk.core.js.primitives import JsObjects


class TeedStreams(object):

  def __init__(self, varId):
    self.varId = varId

  @property
  def stream_0(self):
    return ReadableStream("%s[0]" % self.varId)

  @property
  def stream_1(self):
    return ReadableStream("%s[1]" % self.varId)


class TransformStream(JsPackage):

  @property
  def readable(self):
    """

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/TransformStream

    :return:
    """
    return "%s.readable" % self.varId

  @property
  def writable(self):
    """

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/TransformStream
    """
    return "%s.writable" % self.varId


class ReadableStream(JsPackage):

  @property
  def locked(self):
    return JsObjects.JsPromise("%s.locked" % self.varId)

  def cancel(self):
    """
    Description:
    ------------
    The cancel() method of the ReadableStream interface cancels the associated stream.
    The supplied reason parameter will be given to the underlying source, which may or may not use it.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/ReadableStream/cancel
    """
    return self.fnc("cancel()")

  def getReader(self):
    """
    Description:
    ------------
    The getReader() method of the ReadableStream interface creates a reader and locks the stream to it.
    While the stream is locked, no other reader can be acquired until this one is released.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/ReadableStream/getReader
    """
    return ReadableStream("%s.getReader()" % self.varId)

  def pipeThrough(self, transformStream, options=None):
    """
    Description:
    ------------
    The pipeThrough() method of the ReadableStream interface provides a chainable way of piping the current stream through a transform stream or any other writable/readable pair.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/ReadableStream/pipeThrough

    :param transformStream: A TransformStream (or an object with the structure {writable, readable}) consisting of a readable stream and a writable stream working together to transform some data from one form to another.
    :param options: The options that should be used when piping to the writable stream. Available options are
                      - preventClose
                      - preventAbort
                      - preventCancel
                      - signal
    """
    if options is None:
      return TransformStream("%s.pipeThrough(%s)" % (self.varId, transformStream))

    return TransformStream("%s.pipeThrough(%s, %s)" % (self.varId,transformStream, options))

  def pipeTo(self, destination, options=None):
    """
    Description:
    ------------
    The pipeTo() method of the ReadableStream interface pipes the current ReadableStream to a given WritableStream and returns a promise that fulfills when the piping process completes successfully, or rejects if any errors were encountered.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/ReadableStream/pipeTo

    Attributes:
    ----------
    :param destination:
    :param options:
    """
    if options is None:
      return WritableStream("%s.pipeTo(%s)" % (self.varId, destination))

    return WritableStream("%s.pipeTo(%s, %s)" % (self.varId, destination, options))

  def tee(self):
    """
    Description:
    ------------
    The tee() method of the ReadableStream interface tees the current readable stream, returning a two-element array containing the two resulting branches as new ReadableStream instances.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/ReadableStream/tee
    """
    return TeedStreams("%s.tee()" % self.varId)


class WritableStreamDefaultWriter(JsPackage):

  @property
  def desiredSize(self):
    """
    Description:
    ------------
    The desiredSize read-only property of the WritableStreamDefaultWriter interface returns the desired size required to fill the stream's internal queue.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStreamDefaultWriter/desiredSize
    """
    return "%s.desiredSize" % self.varId

  @property
  def writer(self):
    """
    Description:
    ------------
    The closed read-only property of the WritableStreamDefaultWriter interface returns a promise that fulfills if the stream becomes closed or the writer's lock is released, or rejects if the stream errors.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStreamDefaultWriter/closed
    """
    return JsObjects.JsPromise("%s.writer" % self.varId)

  @property
  def ready(self):
    """
    Description:
    ------------
    The ready read-only property of the WritableStreamDefaultWriter interface returns a Promise that resolves when the desired size of the stream's internal queue transitions from non-positive to positive, signaling that it is no longer applying backpressure.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStreamDefaultWriter/ready
    """
    return JsObjects.JsPromise("%s.ready" % self.varId)

  def abort(self, reason=None):
    """
    Description:
    ------------
    The abort() method of the WritableStreamDefaultWriter interface aborts the stream, signaling that the producer can no longer successfully write to the stream and it is to be immediately moved to an error state, with any queued writes discarded.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStreamDefaultWriter/abort
    """
    if reason is None:
      return JsObjects.JsPromise("%s.abort()" % self.varId)

    reason = JsUtils.jsConvertData(reason, None)
    return JsObjects.JsPromise("%s.abort(%s)" % (self.varId, reason))

  def close(self):
    """
    Description:
    ------------
    The close() method of the WritableStreamDefaultWriter interface closes the associated writable stream.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStreamDefaultWriter/close
    """
    return JsObjects.JsPromise("%s.close()" % self.varId)

  def releaseLock(self):
    """
    Description:
    ------------
    The releaseLock() method of the WritableStreamDefaultWriter interface releases the writer's lock on the corresponding stream.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStreamDefaultWriter/releaseLock
    """
    return self.fnc_closure("%s.releaseLock()" % self.varId)

  def write(self, chunk):
    """
    Description:
    ------------
    The write() property of the WritableStreamDefaultWriter interface writes a passed chunk of data to a WritableStream and its underlying sink, then returns a Promise that resolves to indicate the success or failure of the write operation.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStreamDefaultWriter/write

    Attributes:
    ----------
    :param chunk: A block of binary data to pass to the WritableStream.
    """
    chunk = JsUtils.jsConvertData(chunk, None)
    return JsObjects.JsPromise("%s.write(%s)" % (self.varId, chunk))


class WritableStream(JsPackage):

  @property
  def locked(self):
    """
    Description:
    ------------
    https://developer.mozilla.org/en-US/docs/Web/API/WritableStream
    """
    return "%s.locked" % self.varId

  def abort(self, reason=None):
    """
    Description:
    ------------
    The abort() method of the WritableStream interface aborts the stream, signaling that the producer can no longer successfully write to the stream and it is to be immediately moved to an error state, with any queued writes discarded.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStream/abort
    """
    if reason is None:
      return JsObjects.JsPromise("%s.abort()" % self.varId)

    reason = JsUtils.jsConvertData(reason, None)
    return JsObjects.JsPromise("%s.abort(%s)" % (self.varId, reason))

  def close(self):
    return self.fnc_closure("close()")

  def getWriter(self):
    """
    Description:
    ------------
    The getWriter() method of the WritableStream interface returns a new instance of WritableStreamDefaultWriter and locks the stream to that instance. While the stream is locked, no other writer can be acquired until this one is released.

    Related Pages:

			https://developer.mozilla.org/en-US/docs/Web/API/WritableStream/getWriter
    """
    return WritableStreamDefaultWriter("%s.getWriter()" % self.varId)


if __name__ == '__main__':
  obj = ReadableStream().cancel()
  print(obj.toStr())




from epyk.core.css.styles.classes import CssStyle


class CssHrefNoDecoration(CssStyle.Style):
  _attrs = {'text-decoration': 'none', 'color': 'inherit'}

  def customize(self):
    self.hover.css({'color': self.rptObj.theme.success[1]})


class CssLabelDates(CssStyle.Style):
  _selectors = {'child': 'a'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.colors[5], 'background-image': 'none',
              'color': self.rptObj.theme.greys[0]}, important=True)


class CssHreftMenu(CssStyle.Style):
  _attrs = {'margin': 0, 'clear': 'both', 'border': 0, 'display': 'block', 'position': 'relative', 'color': 'inherit',
            'height': '32px', 'list-style': 'none', 'padding': '0 0 0 5px', 'text-decoration': 'none'}
  _selectors = {'child': 'a'}

  def customize(self):
    self.css({'background': self.rptObj.theme.colors[5], 'color': self.rptObj.theme.colors[4]})


class CssHrefSubMenu(CssStyle.Style):
  _attrs = {'width': '100%', 'padding-left': '5px', 'color': 'white', 'text-decoration': 'none'}
  _selectors = {'child': 'a'}
  
  def customize(self):
    self.css({'color': self.rptObj.theme.greys[-1]})
    self.hover.css({'color': self.rptObj.theme.colors[7]})


class CssSideBarLinks(CssStyle.Style):
  _attrs = {'padding-top': '5px', 'padding-bottom': '5px', 'text-decoration': 'none', 'display': 'block'}
  _hover = {'text-decoration': 'none'}

  def customize(self):
    self.hover.css({'background-color': self.rptObj.theme.greys[2]})


class CssHrefContentLevel1(CssStyle.Style):
  _attrs = {'padding': 0, 'display': 'inline-block', 'margin': 0}


class CssHrefContentLevel2(CssStyle.Style):
  _attrs = {'padding': 0, 'display': 'inline-block', 'margin-left': '20px'}


class CssHrefContentLevel3(CssStyle.Style):
  attrs = {'padding': 0, 'display': 'inline-block', 'margin-left': '40px'}


class CssHrefContentLevel4(CssStyle.Style):
  _attrs = {'padding': 0, 'display': 'inline-block', 'margin-left': '60px'}


class CssFeedbackLink(CssStyle.Style):
  _attrs = {'position': 'fixed', 'bottom': '5px', 'cursor': 'pointer', 'right': '25px', 'padding': '0 10px',
            'z-index': 1000}
  _hover = {'text-decoration': 'underline'}

  def customize(self):
    self.css({'background-color': self.rptObj.theme.greys[2]})


class CssStandardLinks(CssStyle.Style):
  _hover = {'text-decoration': 'underline'}

  def customize(self):
    self.css({'color': self.rptObj.theme.colors[-1]})
    self.hover.css({'color': self.rptObj.theme.colors[-1]})





from epyk.core.html.graph import GraphDC

from epyk.core.js.packages import JsDc


class ChartGeoChoropleth(GraphDC.Chart):
  __reqCss, __reqJs = ['dc'], ['dc', 'crossfilter']

  @property
  def dom(self):
    """
    :rtype: JsChartDC.JsLine
    """
    if self._dom is None:
      self._dom = JsDc.GeoChoropleth(self._report, varName=self.chartId, parent=self)
    return self._dom



"""

"""

from epyk.core.css.catalogs import Catalog

from epyk.core.css.styles.classes import CssStylesList


class CatalogList(Catalog.CatalogGroup):

  def basic(self):
    """  """
    return self._set_class(CssStylesList.CssBasicList)

  def items(self):
    """  """
    return self._set_class(CssStylesList.CssBasicListItems)

  def item_selected(self):
    """ """
    return self._set_class(CssStylesList.CssBasicListItemsSelected)

  def item_no_decoration(self):
    """ """
    return self._set_class(CssStylesList.CssListNoDecoration)

  def item_disabled(self):
    """ """
    return self._set_class(CssStylesList.CssBasicListItemsDisabled)

  def squares(self):
    """  """
    return self._set_class(CssStylesList.CssSquareList)

  def standard(self):
    """  """
    return self._set_class(CssStylesList.CssListBase)

  def items_standard(self):
    """  """
    return self._set_class(CssStylesList.CssListLiBase)

  def sub_list(self):
    """ """
    return self._set_class(CssStylesList.CssListLiUlContainer)

  def sub_item(self):
    """ """
    return self._set_class(CssStylesList.CssListLiSubItem)






import json
import re

from epyk.core.html import Html
from epyk.core.js.Imports import requires

# The list of CSS classes
# from epyk.core.css.styles import GrpCls
# from epyk.core.css.styles import CssGrpClsText


class ExternalLink(Html.Html):
  name, category, callFnc = 'External link', 'Links', 'externallink'

  def __init__(self, report, text, url, icon, helper, height, decoration, options, profile):
    super(ExternalLink, self).__init__(report, {"text": text, "url": url}, css_attrs={'height': height}, profile=profile)
    # Add the internal components icon and helper
    self.add_icon(icon)
    self.add_helper(helper)
    if not 'url' in self.val:
      self.val['url'] = self.val['text']
    if 'target' in options:
      self.set_attrs(name="target", value=options['target'])
    self.decoration, self.options, self.__url = decoration, options, {}

  @property
  def _js__builder__(self):
    return 'htmlObj.innerHTML = data.text; htmlObj.href = data.url'

  def no_decoration(self, color=None):
    """
    Property to remove the list default style
    """
    self.style.css.text_decoration = None
    self.style.list_style_type = None
    if color is None:
      color = self._report.theme.greys[-1]
    self.style.css.color = color
    return self

  def build(self, data=None, options=None, profile=False):
    if not isinstance(data, dict):
      data = {"text": data}
    if "url" not in data:
      data["url"] = self.val["url"]
    return super(ExternalLink, self).build(data, options, profile)

  def __str__(self):
    self.set_attrs(name="href", value=self.val['url'])
    return '<a %s>%s</a>%s' % (self.get_attrs(pyClassNames=self.style.get_classes()), self.val['text'], self.helper)

  # -----------------------------------------------------------------------------------------
  #                                    MARKDOWN SECTION
  # -----------------------------------------------------------------------------------------
  @staticmethod
  def matchMarkDown(val):
    result = re.findall("\[([a-zA-Z 0-9]*)\]\(http([:a-zA-Z\/\.\ 0-9]*)\)", val)
    if not result:
      result = re.findall('\[([a-zA-Z 0-9]*)\]\(<a href\=\"http://([:a-zA-Z/.0-9]*)\">', val)
    return result

  @classmethod
  def convertMarkDown(cls, val, regExpResult, report):
    for name, url in regExpResult:
      val = val.replace("[%s](http%s)" % (name, url), "report.externallink('%s', 'http%s')" % (name, url))
      if report is not None:
        getattr(report, 'link')({'url': url, 'text': name})
    return [val]

  @classmethod
  def jsMarkDown(cls, vals): return "[%s](%s)" % (vals['text'], vals['url'])

  # -----------------------------------------------------------------------------------------
  #                                    EXPORT OPTIONS
  # -----------------------------------------------------------------------------------------
  def to_word(self, document):
    """

    :param document:
    :return:
    """
    pkg_docx = requires("docx", reason='Missing Package', install='python-docx', source_script=__file__)

    p = document.add_paragraph()
    hyperlink = pkg_docx.oxml.shared.OxmlElement('w:hyperlink')
    r_id = p.part.relate_to(self.vals['url'], pkg_docx.opc.constants.RELATIONSHIP_TYPE.HYPERLINK, is_external=True)
    hyperlink.set(pkg_docx.oxml.shared.qn('r:id'), r_id, )
    new_run = pkg_docx.oxml.shared.OxmlElement('w:r')
    new_run.append(pkg_docx.oxml.shared.OxmlElement('w:rPr'))
    new_run.text = self.vals['text']
    hyperlink.append(new_run)
    p._p.append(hyperlink)

  def to_xls(self, workbook, worksheet, cursor):
    worksheet.write_url(cursor['row'], 0, self.vals['url'], string=self.vals['text'])
    cursor['row'] += 2


class DataLink(Html.Html):
  name, category, callFnc = 'Data link', 'Links', 'linkdata'
  # _grpCls = CssGrpClsText.CssClassHref

  def __init__(self, report, text, value, width, height, format, profile):
    super(DataLink, self).__init__(report, {"text": text, 'value': value}, profile=profile,
                                   css_attrs={"width": width, 'height': height})
    self.format = format

  @property
  def no_decoration(self):
    """
    Property to remove the list default style
    """
    self.style.css.text_decoration = None
    self.style.css.list_style_type = None
    return self

  @property
  def _js__builder__(self):
    return '''
      var b = new Blob([data.value]); htmlObj.href = URL.createObjectURL(b);
      htmlObj.innerHTML = data.text'''

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return '<a %(attr)s href="#" download="Download.%(format)s" type="text/%(format)s">%(val)s</a>' % {'attr': self.get_attrs(pyClassNames=self.style.get_classes()), 'val': self.val['text'], 'format': self.format}


class Bridge(Html.Html):
  reqCss, reqJs = [], ['jquery']
  name, category, callFnc = 'Node Bridge', 'Links', 'bridge'

  def __init__(self, report, text, script_name, report_name, url, jsData, context):
    super(Bridge, self).__init__(report, text)
    self.scriptName, self.reportName, self.url = script_name, report_name, url
    if context is not None:
      self.pmts = ["%s: %s" % (k, json.dumps(v))for k, v in context.items()]
    else:
      self.pmts = []
    if jsData is not None:
      for htmlObj in jsData:
        if isinstance(htmlObj, str):
          try:
            htmlObj = self.itemFromCode(htmlObj)
            self.pmts.append("%s: %s" % (htmlObj.htmlCode, htmlObj.val))
          except:
            # hook to assume the user is directly passing a key: value as a string
            self.pmts.append(htmlObj)
        else:
          self.pmts.append("%s: %s" % (htmlObj.htmlCode, htmlObj.val))
    self.addGlobalFnc('NodeBridge(script, report, url, pmts)', '''
      $.post("%s/node/route/" + report + "/" + script, {"url": url}, function(data){
        res = JSON.parse(data);
        if (res.valid){
          var pmt = []; for (k in pmts){pmt.push(k +"="+ pmts[k])};
          if (pmt.length > 0){var url = res['url'] +"?"+ pmt.join("&"); window.open(url, '_blank')}
          else{window.open(res['url'], '_blank')}}}) ''' % self._report._urlsApp['epyk-admin'])

  def __str__(self):
    if self._report.user != "anonymous":
      return "<a %(attr)s onclick='NodeBridge(\"%(scriptName)s\", \"%(reportName)s\", \"%(url)s\", {%(pmts)s})' href='#'>%(vals)s</a>" % {
        'attr': self.get_attrs(pyClassNames=self.defined), 'scriptName': self.scriptName, 'reportName': self.reportName,
        'url': self.url, 'pmts': ",".join(self.pmts), 'vals': self.vals}

    return ""




from epyk.core.html import HtmlList
from epyk.core.html import Html
from epyk.core.html import Defaults

from epyk.core.html.options import OptTrees

from epyk.core.css.styles import GrpClsList


class Tree(HtmlList.List):
  name, category, callFnc = 'List Expandable', 'Lists', 'tree'

  def __init__(self, report, data, color, width, height, htmlCode, helper, option, profile):
    super(Tree, self).__init__(report, [], color, width, height, htmlCode, helper, {}, profile)
    self._vals = data # Attach the original data anyway to the object
    self.__options = OptTrees.OptionsTree(self, option)
    self.set(self, self.val)

  @property
  def options(self):
    """

    :rtype: OptTrees.OptionsTree
    """
    return self.__options

  def empty(self):
    """

    :return:
    """
    return self

  def set(self, ul, data):
    """

    :param ul:
    :param data:
    """
    for l in data:
      if l.get('items') is not None:
        sub_l = self._report.ui.list()
        sub_l.inReport = False
        ul.add_item(sub_l)[-1].no_decoration
        ul[-1].add_label(l.get('label', l.get('value', '')), css={"color": l.get('color', 'none')})
        ul[-1].add_icon(self.options.icon_open if self.options.expanded else self.options.icon_close)
        if not self.options.expanded:
          sub_l.css({"display": 'none'})
        ul[-1].icon.click([
          ul[-1].val.dom.toggle(),
          ul[-1].icon.dom.switchClass(self.options.icon_close.split(" ")[-1], self.options.icon_open.split(" ")[-1])])
        self.set(sub_l, l.get('items'))
      else:
        ul.add_item(l.get('label', l.get('value', '')))[-1].no_decoration
        ul[-1].css({"color": l.get('color', 'none')})
    return self

  def find(self, value):
    pass

  def nodes(self, level):
    """

    :param level:
    :return:
    """
    for i in self.items:
      print(i.items)

  def leafs(self):
    """

    :return:
    """
    pass


class TreeInput(Tree):
  def set(self, ul, data):
    """

    :param ul:
    :param data:
    """
    for l in data:
      if l.get('items') is not None:
        sub_l = self._report.ui.list()
        sub_l.inReport = False
        ul.add_item(sub_l)[-1].no_decoration
        ul[-1].add_label(l['label'], css={"color": l.get('color', 'none')})
        ul[-1].add_icon(self.options.icon_open if self.options.expanded else self.options.icon_close)
        if not self.options.expanded:
          sub_l.css({"display": 'none'})
        ul[-1].icon.click([
          ul[-1].val.dom.toggle(),
          ul[-1].icon.dom.switchClass(self.options.icon_close.split(" ")[-1], self.options.icon_open.split(" ")[-1])])
        self.set(sub_l, l.get('items'))
      else:
        ul.add_item(self._report.ui.text(l['label']).editable().css({"width": 'none', 'min-width': 'none'}))[-1].no_decoration
        ul[-1].css({"color": l.get('color', 'none')})
    return self


class DropDown(Html.Html):
  __reqCss, __reqJs = ['bootstrap'], ['bootstrap', 'jquery']
  #__reqJs = ['jquery']
  name, category, callFnc = 'DropDown Select', 'Lists', 'dropdown'

  def __init__(self, report, data, text, width, height, htmlCode, helper, options, profile):
    self.__options = {}
    super(DropDown, self).__init__(report, text, code=htmlCode, profile=profile, css_attrs={"width": width, "height": height})
    self._vals, self.text = data, text
    self.css({'padding': 0, 'margin': "1px", "display": "block", "z-index": 10, 'cursor': 'pointer', 'position': 'relative'})
    self._jsStyles = {"a": {'text-decoration': 'none', 'line-height': '%spx' % Defaults.LINE_HEIGHT, 'padding': '0 10px', "width": '%spx' % options.get("width")},
                      "ul": {"left": "%spx" % options.get("width")}}
    self.__options = OptTrees.OptDropDown(self, options)

  @property
  def style(self):
    """
    Description:
    -----------

    :rtype: GrpClsList.ClassDropDown
    """
    if self._styleObj is None:
      self._styleObj = GrpClsList.ClassDropDown(self)
    return self._styleObj

  @property
  def options(self):
    """
    Description:
    -----------
    Property to set all the possible object for a Dropdown

    :rtype: OptTrees.OptDropDown
    """
    return self.__options

  @property
  def _js__builder__(self):
    return ''' 
        var jqHtmlObj = jQuery(htmlObj); if(options.clearDropDown) {jqHtmlObj.empty()};
        data.forEach(function(rec){
          if (rec.items != undefined) {
            var li = $('<li class="dropdown" style="list-style-type:none;display:list-item;text-align:-webkit-match-parent"></li>'); var a = $('<a>'+ rec.value +'</a>');
            li.append(a); var ul = $('<ul class="submenu"></ul>'); ul.css(options.ul); options.clearDropDown = false; a.css(options.a);
            %(pyCls)s(ul, rec.items, options); li.append(ul); jqHtmlObj.append(li);
          } else {
              if (rec.url == undefined){var a = $('<a href="#">'+ rec.value +'</a>')}
              else {var a = $('<a href="'+ rec.url +'">'+ rec.value +'</a>')}; a.css(options.a);
              var li = $('<li style="list-style-type:none;display:list-item;text-align:-webkit-match-parent"></li>'); li.append(a); jqHtmlObj.append(li)}
        })''' % {"pyCls": self.__class__.__name__}

  def __str__(self):
    self._report._props.setdefault('js', {}).setdefault("builders", []).append(self.refresh())
    return "<ul %s></ul>" % (self.get_attrs(pyClassNames=self.style.get_classes()))

  def to_word(self, document):
    p = document.add_paragraph()
    p.add_run("Selected: ")
    runner = p.add_run(self._report.http.get(self.htmlCode, self.vals))
    runner.bold = True

  def to_xls(self, workbook, worksheet, cursor):
    if self.htmlId in self._report.http:
      cellTitle = self.title if self.title != "" else 'Input'
      cell_format = workbook.add_format({'bold': True})
      worksheet.write(cursor['row'], 0, cellTitle, cell_format)
      cursor['row'] += 1
      worksheet.write(cursor['row'], 0, self._report.http[self.htmlId])
      cursor['row'] += 2



from epyk.core.html import Defaults


class Meta(object):
  def __init__(self):
    self._metas = {}
    self.__cols = ['charset', 'viewport']
    self.viewport().charset()
    self.http_equiv("X-UA-Compatible", "IE=EDGE").http_equiv("Content-Type", "text/html; charset=utf-8")
    self.http_equiv("Cache-control", "no-cache")

  def viewport(self, value="width=device-width,height=device-height,initial-scale=1.0"):
    """
    Description:
    ------------
    Setting the viewport to make your website look good on all devices

    Usage::

      rptObj.headers.meta.viewport("test")

    Related Pages:

      https://www.w3schools.com/html/html_head.asp

    Attributes:
    ----------
    :param value: Optional. A string with the view port content
    """
    self._metas["viewport"] = '<meta name="viewport" content="%s">' % value
    return self

  def charset(self, value="UTF-8"):
    """
    Description:
    ------------
    Define the character set used

    Usage::

      rptObj.headers.meta.charset("test")

    Related Pages:

			https://www.w3schools.com/tags/tag_meta.asp

    Attributes:
    ----------
    :param value: Optional. A String
    """
    self._metas["charset"] = '<meta charset="%s">' % value
    return self

  def refresh(self, time):
    """
    Description:
    ------------
    Refresh document every X seconds

    Usage::

      rptObj.headers.meta.refresh(10)

    Related Pages:

			https://www.w3schools.com/tags/tag_meta.asp

    Attributes:
    ----------
    :param time: A time in second
    """
    self._metas["refresh"] = '<meta http-equiv="refresh" content="%s">' % time
    if not "refresh" in self.__cols:
      self.__cols.append("refresh")
    return self

  def author(self, name):
    """
    Description:
    ------------
    Define the author of the page

    Usage::

      rptObj.headers.meta.author('epykure')

    Attributes:
    ----------
    :param name: String. The author name
    """
    self._metas["author"] = '<meta name="author" content="%s">' % name
    if not "author" in self.__cols:
      self.__cols.append("author")
    return self

  def description(self, value):
    """
    Description:
    ------------
    Define a description of your web page

    Usage::

      rptObj.headers.meta.description('This is a description')

    Related Pages:

			https://www.w3schools.com/html/html_head.asp

    Attributes:
    ----------
    :param value: String. The report description
    """
    self._metas["description"] = '<meta name="description" content="%s">' % value
    if not "description" in self.__cols:
      self.__cols.append("description")
    return self

  def keywords(self, content):
    """
    Description:
    ------------
    Define keywords for search engine

    Usage::

      rptObj.headers.meta.keywords(['python', 'javascript'])

    Related Pages:

			https://www.w3schools.com/html/html_head.asp

    Attributes:
    ----------
    :param content: String or list with the keyword data
    """
    if isinstance(content, list):
      content = ",".join(content)
    self._metas["keywords"] = '<meta name="keywords" content="%s">' % content
    if not "keywords" in self.__cols:
      self.__cols.append("keywords")
    return self

  def custom(self, name, content):
    """
    Description:
    ------------
    Bespoke function to add other meta tags

    Usage::

      rptObj.headers.meta.custom('language', 'python')

    Attributes:
    ----------
    :param name: String.
    :param content: String.
    """
    self._metas[name] = '<meta name="%s" content="%s">' % (name, content)
    if not name in self.__cols:
      self.__cols.append(name)
    return self

  def http_equiv(self, name, content):
    """
    Description:
    ------------
    Bespoke function to add other http-equiv tags to the meta section

    Usage::

      rptObj.headers.meta.http_equiv('language', 'python')

    Attributes:
    ----------
    :param name: String.
    :param content: String

    :return:
    """
    self._metas[name] = '<meta http-equiv="%s" content="%s">' % (name, content)
    if not name in self.__cols:
      self.__cols.append(name)
    return self

  def __str__(self):
    h = []
    for col in self.__cols:
      if self._metas.get(col) is not None:
        h.append(self._metas.get(col))
    return "\n".join(h)


class Header(object):
  def __init__(self, report=None):
    self._headers, self._links, self._styles, self._scripts, self._base, self.__meta = {}, [], [], [], None, None
    self._favicon_url = Defaults.FAVICON_URL
    if report is not None:
      self._report = report
      self._report._props["header"] = self._headers

  @property
  def meta(self):
    """
    Description:
    ------------
    Property to the Meta data dictionary for the HTML page

    Metadata is data (information) about data.

    The <meta> tag provides metadata about the HTML document.
    Metadata will not be displayed on the page, but will be machine parsable.

    Usage::

      rptObj.headers.meta

    Related Pages:

			https://www.w3schools.com/tags/tag_meta.asp

    :rtype: Meta
    """
    if self.__meta is None:
      self.__meta = Meta()
    return self.__meta

  def title(self, value):
    """
    Description:
    ------------
    The <title> tag is required in all HTML documents and it defines the title of the document.

    Usage::

      rptObj.headers.title("title")

    Related Pages:

			https://www.w3schools.com/tags/tag_title.asp

    Attributes:
    ----------
    :param value: String.
    """
    self._headers['title'] = value
    return self

  def base(self, url):
    """
    Description:
    ------------
    Specify a dedicated path for the relative paths in the page.
    Basically the images will use this path as base if present in the page

    Usage::

      Related Pages:

			https://www.w3schools.com/tags/tag_base.asp

    Attributes:
    ----------
    :param url:

    :return:
    """
    self._base = url
    return self

  def favicon(self, url):
    """
    Description:
    ------------
    The <link> tag defines a link between a document and an external resource.

    The <link> tag is used to link to external style sheets.

    Usage::

      rptObj.headers.favicon('https://github.com/favicon.ico')

    Related Pages:

			https://www.w3schools.com/tags/tag_link.asp

    Attributes:
    ----------
    :param url:

    :return:
    """
    self._favicon_url = url
    return self

  def __str__(self):
    result = [str(self.meta)]
    if self._headers.get("title") is not None:
      result.append("<title>%s</title>" % self._headers.get("title"))
    result.append("<link rel='icon' href='%s' type='image/x-icon'/ >" % self._favicon_url)
    return "\n".join(result)




"""

"""

from epyk.core.js import Js
from epyk.core.js import JsUtils
from epyk.tests import test_statics


# Create a blank Javascript Python object
jsObj = Js.JsBase()

# Write the results to a Javascript file with a Launcher
f = JsUtils.JsFile("DataExamplesNvd3", path=test_statics.OUTPUT_PATHS)

# Load the data from the static sample file
data = test_statics.get_data("flights.txt", n=10)

# Create the data object
record = jsObj.data.records(data)

#record.f.count("B").o.nvd3_bar("count", "column") #.top("count_distinct", n=2, order="asc")
#record.f.count("B").o.nvd3.pie("count", "column") #.top("count_distinct", n=2, order="asc")
#record.f.count("B").o.billboard.line("count", "column") #.top("count_distinct", n=2, order="asc")

data_fnc = record.fnc.count_with_kpi(["origin", "destination"], ["distance", "delay"])

# Write the Javascript fragments to the file
f.writeJs([
  jsObj.console.log(data_fnc), #.a.eq("B", "RR5")),
  #jsObj.console.log(data_fnc.filter.in_("origin", ["MCI", "STL"])), #.a.eq("B", "RR5")),
  #jsObj.console.log(data_fnc.filter.in_("column", "origin")), #.a.eq("B", "RR5")),
  jsObj.console.log(record.to.nvd3.pie(["distance"], "origin")),
])

import json

#data = test_statics.get_data("stocks.json")

# Close the file and print the location of the launcher
print(f.close(jsObj))



from epyk.core import html


class Messaging(object):
  """

  """

  def __init__(self, context):
    self.context = context

  def comments(self, htmlCode, recordset=None, title="", pmts=None, dbService=None, width=(100, '%'),
               height=(200, 'px'), httpCodes=None, readonly=False, profile=None):
    """
    Description:
    ------------
    Python wrapper to a div item composed to several sub html items to display message

    Usage::

      db = report.db(database="test.db")
      report.comments('Test', dbService={'db': db, 'com_table': 'comments', 'reply_table': 'replyComments', 'reply_service': 'post_reply/url', 'user_coms': 'user_comments', 'privacy': 'public', 'service': your/url})

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Comments`

    Related Pages:

        https://leaverou.github.io/bubbly/
    http://manos.malihu.gr/jquery-custom-content-scroller/

    Attributes:
    ----------
    :param htmlCode:
    :param recordset:
    :param title:
    :param pmts:
    :param dbService:
    :param width:
    :param height:
    :param httpCodes:
    :param readonly:
    :param profile:
    :return:
    """
    if recordset is None:
      recordset = []
    return self.context.register(html.HtmlMessaging.Comments(self.context.rptObj, htmlCode, recordset, title, pmts,
                                                             dbService, width, height, httpCodes, readonly, profile))

  def chat(self, htmlCode, title="", pmts=None, dbService=None, width=(100, '%'),
               height=(200, 'px'), httpCodes=None, readonly=False, profile=None, chatOptions=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Chat`

    :param htmlCode:
    :param title:
    :param pmts:
    :param dbService:
    :param width:
    :param height:
    :param httpCodes:
    :param readonly:
    :param profile:
    :param chatOptions:
    :return:
    """
    return self.context.register(html.HtmlMessaging.Chat(self.context.rptObj, htmlCode, title, pmts, dbService, width,
                                                         height, httpCodes, readonly, profile, chatOptions))

  def bot(self, htmlCode, name="Roxane", pmts=None, dbService=None, width=(250, 'px'),
               height=(200, 'px'), httpCodes=None, profile=None, botOptions=None):
    """

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Bot`

    :param htmlCode:
    :param name:
    :param pmts:
    :param dbService:
    :param width:
    :param height:
    :param httpCodes:
    :param profile:
    :param botOptions:
    :return:
    """
    return self.context.register(html.HtmlMessaging.Bot(self.context.rptObj, htmlCode, name, pmts, dbService, width,
                                                        height, httpCodes, profile, botOptions))

  def alert(self, category, title, value, htmlCode=None, background_color=None, close_button=True, width=(320, 'px'),
             height=(None, None), color='black', profile=False, dataSrc=None):
    """
    Description:
    ------------
    Function to add when the python run some tags to put on the top of your report messages.

    The type of the messages can be different according to its criticallity.
    This is fully defined and #driven in the Python and visible in the browser when the page is ready

    All the notification can be hidden directly from the report by setting the flag alerts = False
    e.g: rptObj.alerts = False

    Usage::

      report.ui.messaging.alert('WARNING', 'Server URL not recognized', 'Please check')

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Alert`

    Related Pages:

			https://getbootstrap.com/docs/4.0/components/alerts/

    Attributes:

    :param title: The title of the notification
    :param value: The content of the notification
    :param category:
    :param htmlCode:
    :param background_color: Optional. The background color. Can be also overridden from the .css function
    :param close_button: Optional. Boolean, add a close button to the notification
    :param width:
    :param height:
    :param color:
    :param profile:
    :param dataSrc:
    :rtype: html.HtmlMessaging.Alert
    """
    if not getattr(self.context.rptObj, "alerts", True):
      return

    ui_prop = self.context.rptObj._props.setdefault("ui", {})
    ui_prop.setdefault("notifications", {})["count"] = ui_prop.get("notifications", {}).get("count", 0) + 1
    return self.context.register(html.HtmlMessaging.Alert(self.context.rptObj, title, value, category, width,
                  height, close_button, background_color, color, htmlCode, dataSrc, profile))


  def danger(self, title, value, htmlCode=None, close_button=True, width=(320, 'px'),
             height=(None, None), color='black', profile=False, dataSrc=None):
    """

    Description:
    ------------Function to add when the python run some tags to put on the top of your report messages.

    The type of the messages can be different according to its criticallity.
    This is fully defined and #driven in the Python and visible in the browser when the page is ready

    All the notification can be hidden directly from the report by setting the flag alerts = False
    e.g: rptObj.alerts = False

    Usage::

      rptObj.ui.messaging.alert('danger', 'Server URL not recognized', 'Please check')

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Alert`

    Related Pages:

        https://getbootstrap.com/docs/4.0/components/alerts/

    Attributes:
    ----------
    :param title: The title of the notification
    :param value: The content of the notification
    :param htmlCode:
    :param close_button: Optional. Boolean, add a close button to the notification
    :param width:
    :param height:
    :param color:
    :param profile:
    :param dataSrc:
    :rtype: html.HtmlMessaging.Alert
    """
    if not getattr(self.context.rptObj, "alerts", True):
      return

    ui_prop = self.context.rptObj._props.setdefault("ui", {})
    ui_prop.setdefault("notifications", {})["count"] = ui_prop.get("notifications", {}).get("count", 0) + 1
    return self.context.register(html.HtmlMessaging.Alert(self.context.rptObj, title, value, 'danger', width,
                  height, close_button, self.context.rptObj.getColor('danger', 0), color, htmlCode, dataSrc, profile))

  def info(self, title, value, htmlCode=None, close_button=True, width=(320, 'px'),
           height=(None, None), color='black', profile=False, dataSrc=None):
    """
    Description:
    ------------
    Function to add when the python run some tags to put on the top of your report messages.
    The type of the messages can be different according to its criticallity.
    This is fully defined and #driven in the Python and visible in the browser when the page is ready

    Usage::

      rptObj.ui.messaging.alert('info', 'Server URL not recognized', 'Please check')

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Alert`

    Related Pages:

        https://getbootstrap.com/docs/4.0/components/alerts/

    Attributes:
    ----------
    :rtype: html.HtmlMessaging.Alert
    """
    if not getattr(self.context.rptObj, "alerts", True):
      return

    ui_prop = self.context.rptObj._props.setdefault("ui", {})
    ui_prop.setdefault("notifications", {})["count"] = ui_prop.get("notifications", {}).get("count", 0) + 1
    return self.context.register(html.HtmlMessaging.Alert(self.context.rptObj, title, value, 'info', width,
                  height, close_button, self.context.rptObj.getColor('info', 0), color, htmlCode, dataSrc, profile))

  def success(self, title, value, htmlCode=None, close_button=True, width=(320, 'px'), height=(None, 'black'),
              color='black', profile=False, dataSrc=None):
    """
    Description:
    ------------
    unction to add when the python run some tags to put on the top of your report messages.
    The type of the messages can be different according to its criticallity.
    This is fully defined and #driven in the Python and visible in the browser when the page is ready

    Usage::

      rptObj.ui.messaging.alert('success', 'Server URL not recognized', 'Please check')

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Alert`

    Related Pages:

        https://getbootstrap.com/docs/4.0/components/alerts/

    :param title:
    :param value:
    :param htmlCode:
    :param close_button:
    :param width:
    :param height:
    :param color:
    :param profile:
    :param dataSrc:
    """
    if not getattr(self.context.rptObj, "alerts", True):
      return

    ui_prop = self.context.rptObj._props.setdefault("ui", {})
    ui_prop.setdefault("notifications", {})["count"] = ui_prop.get("notifications", {}).get("count", 0) + 1
    return self.context.register(html.HtmlMessaging.Alert(self.context.rptObj, title, value, 'success', width,
                                                          height, close_button,
                                                          self.context.rptObj.getColor('success', 0), color, htmlCode,
                                                          dataSrc, profile))

  def warning(self, title, value, htmlCode=None, close_button=True, width=(320, 'px'), height=(None, None),
              color='black', profile=False, dataSrc=None):
    """
    Description:
    ------------
    Function to add when the python run some tags to put on the top of your report messages.
    The type of the messages can be different according to its criticallity.
    This is fully defined and #driven in the Python and visible in the browser when the page is ready

    Usage::

      rptObj.ui.messaging.alert('warning', 'Server URL not recognized', 'Please check')

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.Alert`

    Related Pages:

        https://getbootstrap.com/docs/4.0/components/alerts/

    :param title:
    :param value:
    :param htmlCode:
    :param close_button:
    :param width:
    :param height:
    :param color:
    :param profile:
    :param dataSrc:

    """
    if not getattr(self.context.rptObj, "alerts", True):
      return

    ui_prop = self.context.rptObj._props.setdefault("ui", {})
    ui_prop.setdefault("notifications", {})["count"] = ui_prop.get("notifications", {}).get("count", 0) + 1
    return self.context.register(html.HtmlMessaging.Alert(self.context.rptObj, title, value, 'warning', width,
                  height, close_button, self.context.rptObj.getColor('warning', 0), color, htmlCode, dataSrc, profile))

  def news(self, title, value, label=None, link_script=None, icon=None, htmlCode=None, width=(320, 'px'), height=(None, None),
           profile=False):
    """
    Description:
    ------------

    Usage::

      b = rptObj.ui.button("Display")
      n = rptObj.ui.messaging.news("This is a title", "This is the content", link_script="TestSlider")
      b.click(n.jsGenerate("Updated content", isPyData=True))

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlMessaging.News`

    Attributes:
    ----------
    :param title:
    :param value:
    :param label:
    :param link_script:
    :param icon:
    :param htmlCode:
    :param width:
    :param height:
    :param profile:
    """
    return self.context.register(html.HtmlMessaging.News(self.context.rptObj, title, value, label, link_script, icon, width,
                                                          height, htmlCode, profile))




import importlib

from epyk.core import html


class Links(object):
  def __init__(self, context):
    self.context = context

  def external(self, text, url, icon=None, helper=None, height=(None, 'px'), decoration=False, options=None, profile=None):
    """
    Description:
    ------------

    Usage::

      rptObj.ui.links.external('data', 'www.google.fr', icon="fas fa-align-center", options={"target": "_blank"})

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlLinks.ExternalLink`

    Related Pages:

			https://www.w3schools.com/TagS/att_a_href.asp

    Attributes:
    ----------
    :param text: The string value to be displayed in the component
    :param url: The string url of the link
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param helper:
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param decoration:
    :param options: Optional. Specific Python options available for this component
    :param profile: Optional. A flag to set the component performance storage
    """
    dft_options = {"target": '_blank'}
    if options is not None:
      dft_options.update(options)

    #if url.startswith("http"):
    #  url = "/%s" % url
    html_link = html.HtmlLinks.ExternalLink(self.context.rptObj, text, url, icon, helper, height, decoration, dft_options, profile)
    self.context.register(html_link)
    return html_link

  def button(self, text, url, icon=None, helper=None, width=(None, 'px'), height=(None, 'px'), decoration=False, options=None, profile=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlLinks.ExternalLink`

    Attributes:
    ----------
    :param text:
    :param url:
    :param icon:
    :param helper:
    :param height:
    :param decoration:
    :param options:
    :param profile:
    :return:
    """
    if isinstance(width, int):
      width = (width, 'px')
    dft_options = {"target": '_blank'}
    if options is not None:
      dft_options.update(options)
    html_link = html.HtmlLinks.ExternalLink(self.context.rptObj, text, url, icon, helper, height, decoration,
                                            dft_options, profile)
    self.context.register(html_link)
    html_link.style.add_classes.button.basic()
    html_link.style.css.padding = "0 10px"
    return html_link

  def script(self, script_name, report_name=None, icon="fab fa-python", options=None, profile=None):
    """
    Description:
    ------------
    Direct link to another report within the server.

    This is only dedicated to move between reports internally.
    If the target is to move between nodes, the function to use is rather bridge

    Usage::

      rptObj.ui.links.script("TestParams")

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlLinks.ExternalLink`

    Attributes:
    ----------

    :param script_name: The destination script name
    :param report_name: Optional, the report name. Default current one
    :param options: Optional, the link properties

    :rtype: html.HtmlLinks.ExternalLink

    :return:
    """
    script_name = script_name.replace(".py", "")
    text = script_name
    if report_name is None:
      report_name = self.context.rptObj.run.report_name
      mod = importlib.import_module("%s.%s" % (self.context.rptObj.run.report_name, script_name))
      text = mod.TITLE
    if 'name' in options:
      text = options['name']
      del options['name']

    url = "%s/run/%s/%s" % (self.context.rptObj._urlsApp['report'], report_name, script_name)
    return self.link(text, url, icon=icon, options=options, profile=profile)

  def link(self, text="", url="", icon=None, helper=None, height=(None, 'px'), decoration=False, options=None, profile=None):
    """
    Description:
    ------------
    Python interface to the common Hyperlink

    Usage::

      rptObj.ui.link({"text": "Profiling results", "url": '#'})
      l = rptObj.ui.links.link('data', 'www.google.fr', icon="fas fa-align-center", options={"target": "_blank"})
      b = rptObj.ui.images.badge("new")
      l.append_child(b)

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlLinks.ExternalLink`

    Attributes:
    ----------
    :param text: The string value to be displayed in the component
    :param url: The string url of the link
    :param icon: Optional. A string with the value of the icon to display from font-awesome
    :param helper:
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param decoration:
    :param options: Optional. Specific Python options available for this component
    :param profile: Optional. A flag to set the component performance storage
    """
    options = options or {}
    if url is not None and url.startswith("www."):
      url = "//%s" % url
    html_link = html.HtmlLinks.ExternalLink(self.context.rptObj, text, url, icon, helper, height, decoration, options, profile)
    self.context.register(html_link)
    return html_link

  def data(self, text, value, width=(100, '%'), height=(None, 'px'), format='txt', profile=None):
    """
    Description:
    ------------
    Python interface to the Hyperlink to retrieve data

    Usage::

      data_link = rptObj.ui.links.data("link", "test#data")
      data_link.build({"text": 'new link Name', 'data': "new content"})

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlLinks.DataLink`

    Attributes:
    ----------
    :param text: The string value to be displayed in the component
    :param value:
    :param width: Optional. A tuple with the integer for the component width and its unit
    :param height: Optional. A tuple with the integer for the component height and its unit
    :param format: Optional. The downloaded data format
    :param profile: Optional. A flag to set the component performance storage
    """
    html_data = html.HtmlLinks.DataLink(self.context.rptObj, text, value, width=width, height=height, format=format, profile=profile)
    self.context.register(html_data)
    return html_data

  def bridge(self, text, script_name, report_name, url, jsData=None, context=None):
    """
    Description:
    ------------

    Underlying HTML Objects:

      - :class:`epyk.core.html.HtmlLinks.Bridge`

    Related Pages:
Attributes:
    ----------
    :param text:
    :param script_name:
    :param report_name:
    :param url:
    :param jsData:
    :param context:
    """
    html_bridge = html.HtmlLinks.Bridge(self.context.rptObj, text, script_name, report_name, url, jsData, context)
    self.context.register(html_bridge)
    return html_bridge
